import io
import os

import numpy as np
import pytest
import yaml
from numpy.random import random
from numpy.testing import assert_array_equal

import asdf
from asdf import constants, generic_io
from asdf._block import io as bio
from asdf.exceptions import AsdfBlockIndexWarning

RNG = np.random.default_rng(6)


def test_external_block(tmp_path):
    tmp_path = str(tmp_path)

    my_array = RNG.normal(size=(8, 8))
    tree = {"my_array": my_array}
    ff = asdf.AsdfFile(tree)
    ff.set_array_storage(my_array, "external")
    assert ff.get_array_storage(my_array) == "external"

    ff.write_to(os.path.join(tmp_path, "test.asdf"))

    assert "test0000.asdf" in os.listdir(tmp_path)


def test_external_block_url():
    uri = "asdf://foo"
    my_array = RNG.normal(size=(8, 8))
    tree = {"my_array": my_array}
    asdf.get_config().all_array_storage = "external"
    # this should not raise a ValueError since uri is provided
    asdf.AsdfFile(tree, uri=uri)


def test_external_block_non_url():
    my_array = RNG.normal(size=(8, 8))
    tree = {"my_array": my_array}
    ff = asdf.AsdfFile(tree)
    ff.set_array_storage(my_array, "external")
    assert ff.get_array_storage(my_array) == "external"

    buff = io.BytesIO()
    with pytest.raises(ValueError, match=r"Can't write external blocks, since URI of main file is unknown."):
        ff.write_to(buff)


def test_invalid_array_storage():
    my_array = RNG.normal(size=(8, 8))
    tree = {"my_array": my_array}
    ff = asdf.AsdfFile(tree)
    with pytest.raises(ValueError, match=r"array_storage must be one of.*"):
        ff.set_array_storage(my_array, "foo")


def test_transfer_array_sources(tmp_path):
    tmp_path = str(tmp_path)

    my_array = RNG.normal(size=(8, 8))
    tree = {"my_array": my_array}
    ff = asdf.AsdfFile(tree)
    ff.write_to(os.path.join(tmp_path, "test.asdf"))

    with asdf.open(os.path.join(tmp_path, "test.asdf")) as ff:
        assert_array_equal(my_array, ff.tree["my_array"])
        ff.write_to(os.path.join(tmp_path, "test2.asdf"))
        # write_to should have no effect on getting the original data
        assert_array_equal(my_array, ff.tree["my_array"])

    assert ff._fd is None


def test_write_to_same(tmp_path):
    tmp_path = str(tmp_path)

    my_array = RNG.normal(size=(8, 8))
    tree = {"my_array": my_array}
    ff = asdf.AsdfFile(tree)
    ff.write_to(os.path.join(tmp_path, "test.asdf"))

    with asdf.open(os.path.join(tmp_path, "test.asdf"), mode="rw") as ff:
        assert_array_equal(my_array, ff.tree["my_array"])
        ff.tree["extra"] = [0] * 1000
        ff.write_to(os.path.join(tmp_path, "test2.asdf"))

    with asdf.open(os.path.join(tmp_path, "test2.asdf"), mode="rw") as ff:
        assert_array_equal(my_array, ff.tree["my_array"])


def test_pad_blocks(tmp_path):
    tmp_path = str(tmp_path)

    # This is the case where the new tree can't fit in the available space
    my_array = np.ones((8, 8)) * 1
    my_array2 = np.ones((42, 5)) * 2
    tree = {"my_array": my_array, "my_array2": my_array2}

    ff = asdf.AsdfFile(tree)
    ff.write_to(os.path.join(tmp_path, "test.asdf"), pad_blocks=True)

    with asdf.open(os.path.join(tmp_path, "test.asdf")) as ff:
        assert_array_equal(ff.tree["my_array"], my_array)
        assert_array_equal(ff.tree["my_array2"], my_array2)


@pytest.mark.parametrize("lazy_load", [True, False])
@pytest.mark.parametrize("memmap", [True, False])
def test_update_expand_tree(tmp_path, lazy_load, memmap):
    tmp_path = str(tmp_path)
    testpath = os.path.join(tmp_path, "test.asdf")

    # This is the case where the new tree can't fit in the available space
    my_array = np.arange(64) * 1
    my_array2 = np.arange(64) * 2
    tree = {"arrays": [my_array, my_array2, np.arange(3)]}

    ff = asdf.AsdfFile(tree)
    ff.set_array_storage(tree["arrays"][2], "inline")
    ff.write_to(testpath, pad_blocks=True)
    with asdf.open(testpath, lazy_load=lazy_load, memmap=memmap, mode="rw") as ff:
        assert len(list(ff._blocks.blocks)) == 2
        assert_array_equal(ff.tree["arrays"][0], my_array)
        ff.tree["extra"] = [0] * 6000
        ff.update()

    with asdf.open(testpath) as ff:
        assert ff.get_array_storage(ff.tree["arrays"][2]) == "inline"
        assert_array_equal(ff.tree["arrays"][0], my_array)
        assert_array_equal(ff.tree["arrays"][1], my_array2)

    # Now, we expand the header only by a little bit
    ff = asdf.AsdfFile(tree)
    ff.set_array_storage(tree["arrays"][2], "inline")
    ff.write_to(os.path.join(tmp_path, "test2.asdf"), pad_blocks=True)
    with asdf.open(os.path.join(tmp_path, "test2.asdf"), lazy_load=lazy_load, memmap=memmap, mode="rw") as ff:
        ff.tree["extra"] = [0] * 2
        ff.update()

    with asdf.open(os.path.join(tmp_path, "test2.asdf")) as ff:
        assert ff.get_array_storage(ff.tree["arrays"][2]) == "inline"
        assert_array_equal(ff.tree["arrays"][0], my_array)
        assert_array_equal(ff.tree["arrays"][1], my_array2)


@pytest.mark.parametrize("lazy_load", [True, False])
@pytest.mark.parametrize("memmap", [True, False])
def test_update_all_external(tmp_path, lazy_load, memmap):
    fn = tmp_path / "test.asdf"

    my_array = np.arange(64) * 1
    my_array2 = np.arange(64) * 2
    tree = {"arrays": [my_array, my_array2]}

    af = asdf.AsdfFile(tree)
    af.write_to(fn)

    with asdf.config.config_context() as cfg:
        cfg.array_inline_threshold = 10
        cfg.all_array_storage = "external"
        with asdf.open(fn, lazy_load=lazy_load, memmap=memmap, mode="rw") as af:
            af.update()

    assert "test0000.asdf" in os.listdir(tmp_path)
    assert "test0001.asdf" in os.listdir(tmp_path)


@pytest.mark.parametrize("lazy_load", [True, False])
@pytest.mark.parametrize("memmap", [True, False])
def test_update_some_external(tmp_path, lazy_load, memmap):
    fn = tmp_path / "test.asdf"

    my_array = np.arange(64) * 1
    my_array2 = np.arange(64) * 2
    tree = {"arrays": [my_array, my_array2]}

    af = asdf.AsdfFile(tree)
    af.write_to(fn)

    with asdf.open(fn, lazy_load=lazy_load, memmap=memmap, mode="rw") as af:
        af.set_array_storage(af["arrays"][0], "external")
        af.update()

    assert "test0000.asdf" in os.listdir(tmp_path)
    assert "test0001.asdf" not in os.listdir(tmp_path)


def _get_update_tree():
    return {"arrays": [np.arange(64) * 1, np.arange(64) * 2, np.arange(64) * 3]}


@pytest.mark.parametrize("lazy_load", [True, False])
@pytest.mark.parametrize("memmap", [True, False])
def test_update_delete_first_array(tmp_path, lazy_load, memmap):
    tmp_path = str(tmp_path)
    path = os.path.join(tmp_path, "test.asdf")

    # This is the case where the new tree can't fit in the available space
    tree = _get_update_tree()

    ff = asdf.AsdfFile(tree)
    ff.write_to(path, pad_blocks=True)

    original_size = os.stat(path).st_size

    with asdf.open(os.path.join(tmp_path, "test.asdf"), lazy_load=lazy_load, memmap=memmap, mode="rw") as ff:
        del ff.tree["arrays"][0]
        ff.update()

    assert os.stat(path).st_size <= original_size

    with asdf.open(os.path.join(tmp_path, "test.asdf")) as ff:
        assert_array_equal(ff.tree["arrays"][0], tree["arrays"][1])
        assert_array_equal(ff.tree["arrays"][1], tree["arrays"][2])


@pytest.mark.parametrize("lazy_load", [True, False])
@pytest.mark.parametrize("memmap", [True, False])
def test_update_delete_last_array(tmp_path, lazy_load, memmap):
    tmp_path = str(tmp_path)
    path = os.path.join(tmp_path, "test.asdf")

    # This is the case where the new tree can't fit in the available space
    tree = _get_update_tree()

    ff = asdf.AsdfFile(tree)
    ff.write_to(path, pad_blocks=True)

    original_size = os.stat(path).st_size

    with asdf.open(os.path.join(tmp_path, "test.asdf"), lazy_load=lazy_load, memmap=memmap, mode="rw") as ff:
        del ff.tree["arrays"][-1]
        ff.update()

    assert os.stat(path).st_size <= original_size

    with asdf.open(os.path.join(tmp_path, "test.asdf")) as ff:
        assert_array_equal(ff.tree["arrays"][0], tree["arrays"][0])
        assert_array_equal(ff.tree["arrays"][1], tree["arrays"][1])


@pytest.mark.parametrize("lazy_load", [True, False])
@pytest.mark.parametrize("memmap", [True, False])
def test_update_delete_middle_array(tmp_path, lazy_load, memmap):
    tmp_path = str(tmp_path)
    path = os.path.join(tmp_path, "test.asdf")

    # This is the case where the new tree can't fit in the available space
    tree = _get_update_tree()

    ff = asdf.AsdfFile(tree)
    ff.write_to(path, pad_blocks=True)

    original_size = os.stat(path).st_size

    with asdf.open(os.path.join(tmp_path, "test.asdf"), lazy_load=lazy_load, memmap=memmap, mode="rw") as ff:
        del ff.tree["arrays"][1]
        ff.update()
        assert len(ff._blocks.blocks) == 2

    assert os.stat(path).st_size <= original_size

    with asdf.open(os.path.join(tmp_path, "test.asdf")) as ff:
        assert len(ff.tree["arrays"]) == 2
        assert_array_equal(ff.tree["arrays"][0], tree["arrays"][0])
        assert_array_equal(ff.tree["arrays"][1], tree["arrays"][2])


@pytest.mark.parametrize("lazy_load", [True, False])
@pytest.mark.parametrize("memmap", [True, False])
def test_update_replace_first_array(tmp_path, lazy_load, memmap):
    tmp_path = str(tmp_path)
    path = os.path.join(tmp_path, "test.asdf")

    # This is the case where the new tree can't fit in the available space
    tree = _get_update_tree()

    ff = asdf.AsdfFile(tree)
    ff.write_to(path, pad_blocks=True)

    original_size = os.stat(path).st_size

    with asdf.open(os.path.join(tmp_path, "test.asdf"), lazy_load=lazy_load, memmap=memmap, mode="rw") as ff:
        ff.tree["arrays"][0] = np.arange(32)
        ff.update()

    assert os.stat(path).st_size <= original_size

    with asdf.open(os.path.join(tmp_path, "test.asdf")) as ff:
        assert_array_equal(ff.tree["arrays"][0], np.arange(32))
        assert_array_equal(ff.tree["arrays"][1], tree["arrays"][1])
        assert_array_equal(ff.tree["arrays"][2], tree["arrays"][2])


@pytest.mark.parametrize("lazy_load", [True, False])
@pytest.mark.parametrize("memmap", [True, False])
def test_update_replace_last_array(tmp_path, lazy_load, memmap):
    tmp_path = str(tmp_path)
    path = os.path.join(tmp_path, "test.asdf")

    # This is the case where the new tree can't fit in the available space
    tree = _get_update_tree()

    ff = asdf.AsdfFile(tree)
    ff.write_to(path, pad_blocks=True)

    original_size = os.stat(path).st_size

    with asdf.open(os.path.join(tmp_path, "test.asdf"), lazy_load=lazy_load, memmap=memmap, mode="rw") as ff:
        ff.tree["arrays"][2] = np.arange(32)
        ff.update()

    assert os.stat(path).st_size <= original_size

    with asdf.open(os.path.join(tmp_path, "test.asdf")) as ff:
        assert_array_equal(ff.tree["arrays"][0], tree["arrays"][0])
        assert_array_equal(ff.tree["arrays"][1], tree["arrays"][1])
        assert_array_equal(ff.tree["arrays"][2], np.arange(32))


@pytest.mark.parametrize("lazy_load", [True, False])
@pytest.mark.parametrize("memmap", [True, False])
def test_update_replace_middle_array(tmp_path, lazy_load, memmap):
    tmp_path = str(tmp_path)
    path = os.path.join(tmp_path, "test.asdf")

    # This is the case where the new tree can't fit in the available space
    tree = _get_update_tree()

    ff = asdf.AsdfFile(tree)
    ff.write_to(path, pad_blocks=True)

    original_size = os.stat(path).st_size

    with asdf.open(os.path.join(tmp_path, "test.asdf"), lazy_load=lazy_load, memmap=memmap, mode="rw") as ff:
        ff.tree["arrays"][1] = np.arange(32)
        ff.update()

    assert os.stat(path).st_size <= original_size

    with asdf.open(os.path.join(tmp_path, "test.asdf")) as ff:
        assert_array_equal(ff.tree["arrays"][0], tree["arrays"][0])
        assert_array_equal(ff.tree["arrays"][1], np.arange(32))
        assert_array_equal(ff.tree["arrays"][2], tree["arrays"][2])


@pytest.mark.parametrize("lazy_load", [True, False])
@pytest.mark.parametrize("memmap", [True, False])
def test_update_add_array(tmp_path, lazy_load, memmap):
    tmp_path = str(tmp_path)
    path = os.path.join(tmp_path, "test.asdf")

    # This is the case where the new tree can't fit in the available space
    tree = _get_update_tree()

    ff = asdf.AsdfFile(tree)
    ff.write_to(path, pad_blocks=True)

    with asdf.open(os.path.join(tmp_path, "test.asdf"), lazy_load=lazy_load, memmap=memmap, mode="rw") as ff:
        ff.tree["arrays"].append(np.arange(32))
        ff.update()

    with asdf.open(os.path.join(tmp_path, "test.asdf")) as ff:
        assert_array_equal(ff.tree["arrays"][0], tree["arrays"][0])
        assert_array_equal(ff.tree["arrays"][1], tree["arrays"][1])
        assert_array_equal(ff.tree["arrays"][2], tree["arrays"][2])
        assert_array_equal(ff.tree["arrays"][3], np.arange(32))


@pytest.mark.parametrize("lazy_load", [True, False])
@pytest.mark.parametrize("memmap", [True, False])
def test_update_add_array_at_end(tmp_path, lazy_load, memmap):
    tmp_path = str(tmp_path)
    path = os.path.join(tmp_path, "test.asdf")

    # This is the case where the new tree can't fit in the available space
    tree = _get_update_tree()

    ff = asdf.AsdfFile(tree)
    ff.write_to(path, pad_blocks=True)

    original_size = os.stat(path).st_size

    with asdf.open(os.path.join(tmp_path, "test.asdf"), lazy_load=lazy_load, memmap=memmap, mode="rw") as ff:
        ff.tree["arrays"].append(np.arange(65536, dtype="<i8"))
        ff.update()
        assert len(ff._blocks.blocks) == 4

    assert os.stat(path).st_size >= original_size

    with asdf.open(os.path.join(tmp_path, "test.asdf")) as ff:
        assert_array_equal(ff.tree["arrays"][0], tree["arrays"][0])
        assert_array_equal(ff.tree["arrays"][1], tree["arrays"][1])
        assert_array_equal(ff.tree["arrays"][2], tree["arrays"][2])
        assert_array_equal(ff.tree["arrays"][3], np.arange(65536, dtype="<i8"))


@pytest.mark.parametrize("lazy_load", [True, False])
@pytest.mark.parametrize("memmap", [True, False])
def test_update_replace_all_arrays(tmp_path, lazy_load, memmap):
    tmp_path = str(tmp_path)
    testpath = os.path.join(tmp_path, "test.asdf")

    # This is the case where the new tree can't fit in the available space
    my_array = np.ones((64, 64)) * 1
    tree = {
        "my_array": my_array,
    }

    ff = asdf.AsdfFile(tree)
    ff.write_to(testpath, pad_blocks=True)

    with asdf.open(os.path.join(tmp_path, "test.asdf"), lazy_load=lazy_load, memmap=memmap, mode="rw") as ff:
        assert_array_equal(ff.tree["my_array"], np.ones((64, 64)) * 1)
        ff.tree["my_array"] = np.ones((64, 64)) * 2
        ff.update()

    with asdf.open(testpath) as ff:
        assert_array_equal(ff.tree["my_array"], np.ones((64, 64)) * 2)


@pytest.mark.parametrize("lazy_load", [True, False])
@pytest.mark.parametrize("memmap", [True, False])
def test_update_array_in_place(tmp_path, lazy_load, memmap):
    tmp_path = str(tmp_path)
    testpath = os.path.join(tmp_path, "test.asdf")

    # This is the case where the new tree can't fit in the available space
    my_array = np.ones((64, 64)) * 1
    tree = {
        "my_array": my_array,
    }

    ff = asdf.AsdfFile(tree)
    ff.write_to(testpath, pad_blocks=True)

    with asdf.open(testpath, lazy_load=lazy_load, memmap=memmap, mode="rw") as ff:
        array = np.asarray(ff.tree["my_array"])
        array *= 2
        ff.update()

    with asdf.open(testpath) as ff:
        assert_array_equal(ff.tree["my_array"], np.ones((64, 64)) * 2)


@pytest.mark.parametrize("lazy_load", [True, False])
@pytest.mark.parametrize("memmap", [True, False])
def test_update_compressed_blocks(tmp_path, lazy_load, memmap):
    """
    This test was originally constructed to test an issue where
    a failed update left a corrupt file. The issue that resulted in
    the failed update (a compressed block growing in size) was fixed
    so this is no longer a good test for a failed update.

    See: https://github.com/asdf-format/asdf/issues/1520

    However, the test does serve to make sure that updating the
    contents of compressed blocks in a way that causes them to grow
    in size on disk does not result in a failed update.
    """
    fn = tmp_path / "test.asdf"
    n_arrays = 10
    array_size = 10000

    # make a tree with many arrays that will compress well
    af = asdf.AsdfFile()
    for i in range(n_arrays):
        af[i] = np.zeros(array_size, dtype="uint8") + i
        af.set_array_compression(af[i], "zlib")
    af.write_to(fn)

    with asdf.open(fn, lazy_load=lazy_load, memmap=memmap, mode="rw") as af:
        # now make the data are difficult to compress
        for i in range(n_arrays):
            assert np.all(af[i] == i)
            af[i][:] = np.random.randint(255, size=array_size)
            af[i][0] = i + 1
        af.update()

    with asdf.open(fn, mode="r") as af:
        for i in range(n_arrays):
            assert af[i][0] == i + 1


def test_init_from_asdffile(tmp_path):
    tmp_path = str(tmp_path)

    my_array = RNG.normal(size=(8, 8))
    tree = {"my_array": my_array}
    ff = asdf.AsdfFile(tree)
    ff2 = asdf.AsdfFile(ff)
    assert ff.tree["my_array"] is ff2.tree["my_array"]
    assert_array_equal(ff.tree["my_array"], ff2.tree["my_array"])

    ff2.tree["my_array"] = None
    assert_array_equal(ff.tree["my_array"], my_array)

    ff.write_to(os.path.join(tmp_path, "test.asdf"))

    with asdf.open(os.path.join(tmp_path, "test.asdf")) as ff:
        ff2 = asdf.AsdfFile(ff)
        # assert ff.tree["my_array"] is not ff2.tree["my_array"]
        assert_array_equal(ff.tree["my_array"], ff2.tree["my_array"])

        ff2.tree["my_array"] = None
        assert_array_equal(ff.tree["my_array"], my_array)


def test_seek_until_on_block_boundary():
    # Create content where the first block begins on a
    # file-reading-block boundary.

    content = b"""#ASDF 1.0.0
%YAML 1.1
%TAG ! tag:stsci.edu:asdf/
--- !core/asdf-1.0.0
foo : bar
...
"""
    content += b"\0" * (io.DEFAULT_BUFFER_SIZE - 2) + constants.BLOCK_MAGIC + b"\0\x30" + b"\0" * 48

    buff = io.BytesIO(content)
    ff = asdf.open(buff)
    assert len(ff._blocks.blocks) == 1

    buff.seek(0)
    fd = generic_io.InputStream(buff, "r")
    ff = asdf.open(fd)
    assert len(ff._blocks.blocks) == 1


@pytest.mark.parametrize("write_checksums", [True, False])
def test_checksum(tmp_path, write_checksums):
    tmp_path = str(tmp_path)
    path = os.path.join(tmp_path, "test.asdf")

    my_array = np.arange(0, 64, dtype="<i8").reshape((8, 8))
    tree = {"my_array": my_array}
    ff = asdf.AsdfFile(tree)
    ff.write_to(path, write_checksums=write_checksums)

    with asdf.open(path, validate_checksums=True) as ff:
        if write_checksums:
            checksum = b"\xcaM\\\xb8t_L|\x00\n+\x01\xf1\xcfP1"
        else:
            checksum = b"\0" * 16

        assert ff._blocks.blocks[0].header["checksum"] == checksum


@pytest.mark.parametrize("lazy_load", [True, False])
@pytest.mark.parametrize("memmap", [True, False])
@pytest.mark.parametrize("write_checksums", [True, False])
def test_checksum_update(tmp_path, lazy_load, memmap, write_checksums):
    tmp_path = str(tmp_path)
    path = os.path.join(tmp_path, "test.asdf")

    my_array = np.arange(0, 64, dtype="<i8").reshape((8, 8))

    tree = {"my_array": my_array}
    ff = asdf.AsdfFile(tree)
    ff.write_to(path, write_checksums=write_checksums)

    with asdf.open(path, lazy_load=lazy_load, memmap=memmap, mode="rw") as ff:
        ff.tree["my_array"][7, 7] = 0.0
        # update() should update the checksum, even if the data itself
        # is memmapped and isn't expressly re-written.
        ff.update(write_checksums=write_checksums)

    with asdf.open(path, validate_checksums=True) as ff:
        if write_checksums:
            checksum = b"T\xaf~[\x90\x8a\x88^\xc2B\x96D,N\xadL"
        else:
            checksum = b"\0" * 16

        assert ff._blocks.blocks[0].header["checksum"] == checksum


def test_block_index():
    buff = io.BytesIO()

    arrays = []
    for i in range(100):
        arrays.append(np.ones((8, 8)) * i)

    tree = {"arrays": arrays}

    ff = asdf.AsdfFile(tree)
    ff.write_to(buff)

    buff.seek(0)
    with asdf.open(buff) as ff2:
        assert len(ff2._blocks.blocks) == 100
        assert ff2._blocks.blocks[0].loaded
        for i in range(2, 99):
            assert not ff2._blocks.blocks[i].loaded
        assert ff2._blocks.blocks[99].loaded

        # Force the loading of one array
        ff2.tree["arrays"][50] * 2

        for i in range(2, 99):
            if i == 50:
                assert ff2._blocks.blocks[i].loaded
            else:
                assert not ff2._blocks.blocks[i].loaded


def test_large_block_index():
    """
    This test is designed to test reading of a block index that is
    larger than a single file system block, which is why we create
    io.DEFAULT_BUFFER_SIZE / 4 arrays, and assuming each entry has more
    than one digit in its address, we're guaranteed to have an index
    larger than a filesystem block.
    """

    # TODO: It would be nice to find a way to make this test faster.  The
    # real bottleneck here is the enormous YAML section.

    buff = io.BytesIO()

    narrays = int(io.DEFAULT_BUFFER_SIZE / 4)

    arrays = []
    for i in range(narrays):
        arrays.append(np.array([i], np.uint16))

    tree = {"arrays": arrays}

    ff = asdf.AsdfFile(tree)
    # Since we're testing with small arrays, force all arrays to be stored
    # in internal blocks rather than letting some of them be automatically put
    # inline.
    ff.write_to(buff, all_array_storage="internal")

    buff.seek(0)
    with asdf.open(buff) as ff2:
        assert ff2._blocks.blocks[0].loaded
        assert len(ff2._blocks.blocks) == narrays


def test_no_block_index():
    buff = io.BytesIO()

    arrays = []
    for i in range(10):
        arrays.append(np.ones((8, 8)) * i)

    tree = {"arrays": arrays}

    ff = asdf.AsdfFile(tree)
    ff.write_to(buff, include_block_index=False)

    assert constants.INDEX_HEADER not in buff.getvalue()


def test_junk_after_index():
    buff = io.BytesIO()

    arrays = []
    for i in range(10):
        arrays.append(np.ones((8, 8)) * i)

    tree = {"arrays": arrays}

    ff = asdf.AsdfFile(tree)
    ff.write_to(buff)

    buff.write(b"JUNK")

    buff.seek(0)

    # This has junk after the block index, so it
    # should fall back to reading serially
    with pytest.warns(AsdfBlockIndexWarning, match="Failed to read block index"):
        with asdf.open(buff) as ff:
            assert ff._blocks.blocks[1].loaded


def test_short_file_find_block_index():
    # This tests searching for a block index in a file that looks like
    # it might have an index, in the last filesystem block or so, but
    # ultimately proves to not have an index.

    buff = io.BytesIO()

    ff = asdf.AsdfFile({"arr": np.ndarray([1]), "arr2": np.ndarray([2])})
    # Since we're testing with small arrays, force all arrays to be stored
    # in internal blocks rather than letting some of them be automatically put
    # inline.
    ff.write_to(buff, include_block_index=False, all_array_storage="internal")

    buff.write(b"#ASDF BLOCK INDEX\n")
    buff.write(b"0" * (io.DEFAULT_BUFFER_SIZE * 4))

    buff.seek(0)
    with pytest.warns(AsdfBlockIndexWarning, match="Failed to read block index"):
        with asdf.open(buff) as ff:
            assert len(ff._blocks.blocks) == 2
            assert ff._blocks.blocks[1].loaded


def test_invalid_block_index_values():
    # This adds a value in the block index that points to something
    # past the end of the file.  In that case, we should just reject
    # the index altogether.

    buff = generic_io.get_file(io.BytesIO(), mode="w")

    arrays = []
    for i in range(10):
        arrays.append(np.ones((8, 8)) * i)

    tree = {"arrays": arrays}

    ff = asdf.AsdfFile(tree)
    ff.write_to(buff, include_block_index=True)
    buff.seek(0)
    offset = bio.find_block_index(buff)
    buff.seek(offset)
    block_index = bio.read_block_index(buff)
    block_index.append(123456789)
    buff.seek(offset)
    bio.write_block_index(buff, block_index)

    buff.seek(0)
    with pytest.warns(AsdfBlockIndexWarning, match="Invalid block index contents"):
        with asdf.open(buff) as ff:
            assert len(ff._blocks.blocks) == 10
            assert ff._blocks.blocks[1].loaded


@pytest.mark.parametrize("block_index_index", [0, -1])
def test_invalid_block_index_offset(block_index_index):
    """
    This adds a value in the block index that points to something
    that isn't a block
    """

    buff = io.BytesIO()

    arrays = []
    for i in range(10):
        arrays.append(np.ones((8, 8)) * i)

    tree = {"arrays": arrays}

    ff = asdf.AsdfFile(tree)
    ff.write_to(buff)

    # now overwrite the block index with the first entry
    # incorrectly pointing to a non-block offset
    buff.seek(0)
    bs = buff.read()
    block_index_header_start = bs.index(constants.INDEX_HEADER)
    block_index_start = block_index_header_start + len(constants.INDEX_HEADER)
    block_index = yaml.load(bs[block_index_start:], yaml.SafeLoader)
    block_index[block_index_index] -= 4
    buff.seek(block_index_start)
    yaml.dump(
        block_index,
        stream=buff,
        explicit_start=True,
        explicit_end=True,
        version=asdf.versioning._YAML_VERSION,
        allow_unicode=True,
        encoding="utf-8",
    )

    buff.seek(0)
    with pytest.warns(AsdfBlockIndexWarning, match="Invalid block index contents"):
        with asdf.open(buff) as ff:
            assert len(ff._blocks.blocks) == 10
            for i, a in enumerate(arrays):
                assert ff._blocks.blocks[i].loaded
                assert_array_equal(ff["arrays"][i], a)


def test_unordered_block_index():
    """
    This creates a block index that isn't in increasing order
    """

    buff = generic_io.get_file(io.BytesIO(), mode="w")

    arrays = []
    for i in range(10):
        arrays.append(np.ones((8, 8)) * i)

    tree = {"arrays": arrays}

    ff = asdf.AsdfFile(tree)
    ff.write_to(buff, include_block_index=True)
    buff.seek(0)
    offset = bio.find_block_index(buff)
    buff.seek(offset)
    block_index = bio.read_block_index(buff)
    buff.seek(offset)
    bio.write_block_index(buff, block_index[::-1])
    buff.seek(0)

    buff.seek(0)
    with pytest.warns(AsdfBlockIndexWarning, match="Failed to read block index"):
        with asdf.open(buff) as ff:
            assert len(ff._blocks.blocks) == 10
            assert ff._blocks.blocks[1].loaded


@pytest.fixture(scope="module")
def filename_with_array(tmp_path_factory):
    fn = tmp_path_factory.mktemp("data") / "filename_with_array.asdf"
    tree = {"array": np.random.random((20, 20))}
    ff = asdf.AsdfFile(tree)
    ff.write_to(fn)
    return fn


@pytest.mark.parametrize(
    "open_kwargs,should_memmap",
    [
        ({}, False),
        ({"lazy_load": True, "memmap": True}, True),
        ({"lazy_load": False, "memmap": True}, True),
        ({"lazy_load": True, "memmap": False}, False),
        ({"lazy_load": False, "memmap": False}, False),
    ],
)
def test_open_no_memmap(filename_with_array, open_kwargs, should_memmap):
    """
    Test that asdf.open does not (or does) return memmaps for arrays
    depending on a number of arguments including:
        default (no kwargs)
        memmap
    """
    with asdf.open(filename_with_array, **open_kwargs) as af:
        array = af.tree["array"]
        if should_memmap:
            assert isinstance(array.base, np.memmap)
        else:
            assert not isinstance(array.base, np.memmap)


def test_add_block_before_fully_loaded(tmp_path):
    """
    This test covers a subtle case where a block is added
    to a file before all pre-existing internal blocks have
    been located.  If the BlockManager isn't careful to
    locate them all first, the new block will take the index
    of an existing block and views over that index will
    point to the wrong data.

    See https://github.com/asdf-format/asdf/issues/999
    """
    file_path1 = tmp_path / "test1.asdf"
    file_path2 = tmp_path / "test2.asdf"
    arr0 = random(10)
    arr1 = random(10)
    arr2 = random(10)

    with asdf.AsdfFile() as af:
        af["arr0"] = None
        af["arr1"] = arr1
        af["arr2"] = arr2
        af.write_to(file_path1, include_block_index=False)

    with asdf.open(file_path1) as af:
        af["arr0"] = arr0
        af.write_to(file_path2)

    with asdf.open(file_path2) as af:
        assert_array_equal(af["arr0"], arr0)
        assert_array_equal(af["arr1"], arr1)
        assert_array_equal(af["arr2"], arr2)


@pytest.mark.parametrize("all_array_storage", ["internal", "external", "inline"])
@pytest.mark.parametrize("all_array_compression", [None, "", "zlib", "bzp2", "lz4", "input"])
@pytest.mark.parametrize("compression_kwargs", [None, {}])
def test_write_to_update_storage_options(tmp_path, all_array_storage, all_array_compression, compression_kwargs):
    if all_array_compression == "bzp2" and compression_kwargs is not None:
        compression_kwargs = {"compresslevel": 1}

    def assert_result(ff):
        if all_array_storage == "external":
            assert "test0000.asdf" in os.listdir(tmp_path)
        else:
            assert "test0000.asdf" not in os.listdir(tmp_path)
        if all_array_storage == "internal":
            assert len(ff._blocks.blocks) == 1
        else:
            assert len(ff._blocks.blocks) == 0

        if all_array_storage == "internal":
            target_compression = all_array_compression or None
            if target_compression == "input":
                target_compression = None
            assert ff.get_array_compression(ff["array"]) == target_compression

    arr1 = np.ones((8, 8))
    tree = {"array": arr1}
    fn = tmp_path / "test.asdf"

    ff1 = asdf.AsdfFile(tree)

    # first check write_to
    ff1.write_to(
        fn,
        all_array_storage=all_array_storage,
        all_array_compression=all_array_compression,
        compression_kwargs=compression_kwargs,
    )

    # then reuse the file to check update
    arr2 = np.ones((8, 8)) * 42
    with asdf.open(fn, mode="rw") as ff2:
        assert_result(ff2)
        np.testing.assert_array_equal(arr1, ff2["array"])
        ff2["array"] = arr2
        ff2.update(
            all_array_storage=all_array_storage,
            all_array_compression=all_array_compression,
            compression_kwargs=compression_kwargs,
        )
        assert_result(ff2)
    with asdf.open(fn) as ff3:
        assert_result(ff3)
        np.testing.assert_array_equal(arr2, ff3["array"])


@pytest.mark.parametrize("lazy_load", [True, False])
@pytest.mark.parametrize("memmap", [True, False])
def test_remove_blocks(tmp_path, lazy_load, memmap):
    fn1 = tmp_path / "test.asdf"
    fn2 = tmp_path / "test2.asdf"

    tree = {"a": np.zeros(3), "b": np.ones(10)}
    tree["c"] = tree["b"][:5]

    for key in tree:
        af = asdf.AsdfFile(tree)
        af.write_to(fn1)

        with asdf.open(fn1, lazy_load=lazy_load, memmap=memmap, mode="rw") as af:
            assert len(af._blocks.blocks) == 2
            af[key] = None
            af.write_to(fn2)

        with asdf.open(fn1, lazy_load=lazy_load, memmap=memmap, mode="rw") as af:
            assert len(af._blocks.blocks) == 2
            af[key] = None
            af.update()

        for fn in (fn1, fn2):
            with asdf.open(fn) as af:
                if key == "a":
                    assert len(af._blocks.blocks) == 1
                else:
                    assert len(af._blocks.blocks) == 2
                for key2 in tree:
                    if key == key2:
                        continue
                    np.testing.assert_array_equal(af[key2], tree[key2])


def test_open_memmap_from_closed_file(tmp_path):
    fn = tmp_path / "test.asdf"
    arr = np.zeros(100)
    arr2 = np.ones(100)
    tree = {"base": arr, "view": arr[:50], "base2": arr2}
    af = asdf.AsdfFile(tree)
    af.write_to(fn)

    with asdf.open(fn, lazy_load=True, memmap=False) as af:
        # load the base so we can test if accessing the view after the
        # file is closed will trigger an error
        af["base"][:]
        view = af["view"]
        base2 = af["base2"]

    msg = r"ASDF file has already been closed. Can not get the data."
    with pytest.raises(OSError, match=msg):
        view[:]

    with pytest.raises(OSError, match=msg):
        base2[:]


@pytest.mark.parametrize("default_array_save_base", [True, False])
@pytest.mark.parametrize("save_base", [True, False, None])
def test_views_save_base(tmp_path, default_array_save_base, save_base):
    fn = tmp_path / "test.asdf"
    arr = np.zeros(100, dtype="uint8")
    tree = {"v": arr[:10]}
    with asdf.config_context() as cfg:
        cfg.default_array_save_base = default_array_save_base
        af = asdf.AsdfFile(tree)
        if save_base is not None:
            af.set_array_save_base(af["v"], save_base)
        af.write_to(fn)

    with asdf.open(fn, memmap=True) as af:
        base = af["v"].base
        if save_base or (save_base is None and default_array_save_base):
            assert len(base) == 100
        else:
            assert len(base) == 10



"""
Commands for displaying summaries of ASDF trees
"""

import asdf

from .main import Command

__all__ = ["info"]


class Info(Command):
    @classmethod
    def setup_arguments(cls, subparsers):
        parser = subparsers.add_parser(
            "info",
            help="Print a rendering of an ASDF tree.",
            description="""Show the contents of an ASDF file by rendering tree contents as text.
            For large files max-rows and max-cols can be used to truncate the output to improve
            readability. When a file contains more lines to display than max-rows the
            deepest parts of the tree will be hidden with a message saying how many
            rows are hidden.""",
        )

        parser.add_argument("filename", help="ASDF file to render")
        parser.add_argument(
            "--max-rows", type=int, help="Maximum number of lines to print. If not provided all lines will be shown."
        )
        parser.add_argument(
            "--max-cols", type=int, help="Maximum length of line. If not provided lines will have no length limit."
        )

        parser.add_argument(
            "--show-values",
            dest="show_values",
            action="store_true",
            help="Display primitive values in the rendered tree (by default, enabled).",
        )
        parser.add_argument("--no-show-values", dest="show_values", action="store_false")
        parser.set_defaults(show_values=True)
        parser.add_argument(
            "--show-blocks",
            action="store_true",
            help="Display a table of header information for each block in the file.",
        )

        parser.set_defaults(func=cls.run)

        return parser

    @classmethod
    def run(cls, args):
        info(args.filename, args.max_rows, args.max_cols, args.show_values, args.show_blocks)


def info(filename, max_rows, max_cols, show_values, show_blocks):
    with asdf.open(filename) as af:
        af.info(max_rows, max_cols, show_values, show_blocks=show_blocks)



import contextlib
import io
import mmap
import os

import numpy as np
import pytest

from asdf import constants, generic_io, util
from asdf._block import io as bio
from asdf._block.reader import read_blocks
from asdf.exceptions import AsdfBlockIndexWarning, AsdfWarning


@contextlib.contextmanager
def gen_blocks(
    fn=None,
    n=5,
    size=10,
    padding=0,
    padding_byte=b"\0",
    with_index=False,
    block_padding=False,
    streamed=False,
    write_checksums=False,
):
    offsets = []
    if fn is not None:
        with generic_io.get_file(fn, mode="w") as fd:
            pass

    def check(blocks):
        assert len(blocks) == n
        for i, blk in enumerate(blocks):
            assert blk.data.size == size
            assert np.all(blk.data == i)

    with generic_io.get_file(fn or io.BytesIO(), mode="rw") as fd:
        fd.write(padding_byte * padding)
        for i in range(n):
            offsets.append(fd.tell())
            fd.write(constants.BLOCK_MAGIC)
            data = np.ones(size, dtype="uint8") * i
            bio.write_block(
                fd, data, stream=streamed and (i == n - 1), padding=block_padding, write_checksum=write_checksums
            )
        if with_index and not streamed:
            bio.write_block_index(fd, offsets)
        fd.seek(0)
        yield fd, check


# test a few paddings to test read_blocks checking 4 bytes while searching for the first block
@pytest.mark.parametrize("lazy_load", [True, False])
@pytest.mark.parametrize("memmap", [True, False])
@pytest.mark.parametrize("with_index", [True, False])
@pytest.mark.parametrize("validate_checksums", [True, False])
@pytest.mark.parametrize("padding", [0, 3, 4, 5])
@pytest.mark.parametrize("streamed", [True, False])
@pytest.mark.parametrize("write_checksums", [True, False])
def test_read(tmp_path, lazy_load, memmap, with_index, validate_checksums, padding, streamed, write_checksums):
    fn = tmp_path / "test.bin"
    n = 5
    size = 10
    with gen_blocks(
        fn=fn,
        n=n,
        size=size,
        padding=padding,
        with_index=with_index,
        streamed=streamed,
        write_checksums=write_checksums,
    ) as (fd, check):
        r = read_blocks(fd, memmap=memmap, lazy_load=lazy_load, validate_checksums=validate_checksums)
        if lazy_load and with_index and not streamed:
            assert r[0].loaded
            assert r[-1].loaded
            for blk in r[1:-1]:
                assert not blk.loaded
                # getting the header should load the block
                blk.header
                assert blk.loaded
        else:
            for blk in r:
                assert blk.loaded
        if memmap:
            for blk in r:
                base = util.get_array_base(blk.data)
                assert isinstance(base.base, mmap.mmap)
        check(r)
        if lazy_load:
            # if lazy loaded, each call to data should re-read the data
            assert r[0].data is not r[0].data
        else:
            assert r[0].data is r[0].data
        # getting cached_data should always return the same array
        assert r[0].cached_data is r[0].cached_data


@pytest.mark.parametrize("padding", (1, 4, 7))
@pytest.mark.parametrize("padding_byte", (b"\1", b"\0", b" ", b"\xd3", b"B", b"L", b"K", b"\xd3BL"))
def test_read_valid_padding(padding, padding_byte):
    """Test that reader allows padding bytes before the first block"""
    with gen_blocks(padding=padding, padding_byte=padding_byte) as (fd, check):
        check(read_blocks(fd))


@pytest.mark.parametrize("padding_byte", (b"\xd3BLK", b" \xd3BLK"))
def test_read_invalid_padding(padding_byte):
    with gen_blocks(padding=1, padding_byte=padding_byte) as (fd, check):
        with pytest.raises(ValueError, match="buffer is smaller than requested size"):
            check(read_blocks(fd))


def test_read_post_padding_null_bytes():
    with gen_blocks(padding=1) as (fd, check):
        fd.seek(0, os.SEEK_END)
        # acceptable to have <4 bytes after the last block
        fd.write(b"\x00" * 3)
        fd.seek(0)
        check(read_blocks(fd))


def test_read_post_padding_non_null_bytes():
    with gen_blocks(padding=1) as (fd, check):
        fd.seek(0, os.SEEK_END)
        # acceptable to have <4 bytes after the last block
        fd.write(b"\x01" * 3)
        fd.seek(0)
        with pytest.warns(AsdfWarning, match=r"Read invalid bytes.*"):
            check(read_blocks(fd))


@pytest.mark.parametrize("invalid_block_index", [0, 1, -1, "junk"])
def test_invalid_block_index(tmp_path, invalid_block_index):
    fn = tmp_path / "test.bin"
    with gen_blocks(fn=fn, with_index=True) as (fd, check):
        # trash the block index
        offset = bio.find_block_index(fd)
        assert offset is not None
        if invalid_block_index == "junk":
            # trash the whole index
            fd.seek(-4, 2)
            fd.write(b"junk")
        else:  # mess up one entry of the index
            block_index = bio.read_block_index(fd, offset)
            block_index[invalid_block_index] += 4
            fd.seek(offset)
            bio.write_block_index(fd, block_index)
        fd.seek(0)

        # when the block index is read, only the first and last blocks
        # are check, so any other invalid entry should result in failure
        if invalid_block_index in (0, -1):
            with pytest.warns(AsdfBlockIndexWarning, match="Invalid block index contents"):
                check(read_blocks(fd, lazy_load=True))
        elif invalid_block_index == "junk":
            # read_blocks should fall back to reading serially
            with pytest.warns(AsdfBlockIndexWarning, match="Failed to read block index"):
                check(read_blocks(fd, lazy_load=True))
        else:
            with pytest.raises(ValueError, match=r"Header size.*"):
                check(read_blocks(fd, lazy_load=True))


def test_invalid_block_in_index_with_valid_magic(tmp_path):
    fn = tmp_path / "test.bin"
    with gen_blocks(fn=fn, with_index=True, block_padding=1.0) as (fd, check):
        offset = bio.find_block_index(fd)
        assert offset is not None
        block_index = bio.read_block_index(fd, offset)
        # move the first block offset to the padding before
        # the second block with enough space to write
        # valid magic (but invalid header)
        block_index[0] = block_index[1] - 6
        fd.seek(block_index[0])
        fd.write(constants.BLOCK_MAGIC)
        fd.write(b"\0\0")

        fd.seek(offset)
        bio.write_block_index(fd, block_index)

        fd.seek(0)
        with pytest.warns(AsdfBlockIndexWarning, match="Invalid block index contents"):
            check(read_blocks(fd, lazy_load=True))


def test_closed_file(tmp_path):
    fn = tmp_path / "test.bin"
    with gen_blocks(fn=fn, with_index=True) as (fd, _):
        blocks = read_blocks(fd, lazy_load=True)
        blk = blocks[1]
    with pytest.raises(OSError, match="Attempt to load block from closed file"):
        blk.load()


def empty_header(checksum, compression=b"\0\0\0\0"):
    """Return a header with all fields other than checksum set to zero."""

    return io.BytesIO(
        constants.BLOCK_MAGIC
        + b"\x000"  # header size = 2
        + b"\0\0\0\0"  # flags = 4
        + compression  # compression = 4
        + b"\0\0\0\0\0\0\0\0"  # allocated size = 8
        + b"\0\0\0\0\0\0\0\0"  # used size = 8
        + b"\0\0\0\0\0\0\0\0"  # data size = 8
        + checksum  # checksum = 16
    )


@pytest.mark.parametrize("validate_checksums", [True, False])
def test_missing_checksum(validate_checksums):
    buff = empty_header(b"\0" * 16)
    with generic_io.get_file(buff, mode="r") as fd:
        # All-zero checksum should always be allowed
        read_blocks(fd, lazy_load=False, validate_checksums=validate_checksums)[0].data


@pytest.mark.parametrize("validate_checksums", [True, False])
@pytest.mark.parametrize("compression", [b"\0\0\0\0", b"lz4\0"])
def test_bad_checksum(validate_checksums, compression):
    buff = empty_header(b"\1" + b"\0" * 15, compression)
    with generic_io.get_file(buff, mode="r") as fd:
        if validate_checksums:
            # Invalid checksum should raise an exception if `validate_checksums=True`
            with pytest.raises(ValueError, match=r".* does not match given checksum"):
                block = read_blocks(fd, lazy_load=False, validate_checksums=validate_checksums)[0]
                _ = block.data
        else:
            block = read_blocks(fd, lazy_load=False, validate_checksums=validate_checksums)[0]
            _ = block.data



import contextlib
import gc

import numpy as np
from numpy.testing import assert_array_equal

import asdf
from asdf.extension import Converter, Extension
from asdf.testing import helpers


class BlockData:
    def __init__(self, payload):
        self.payload = payload


class BlockConverter(Converter):
    tags = ["asdf://somewhere.org/tags/block_data-1.0.0"]
    types = [BlockData]
    _return_invalid_keys = False

    def to_yaml_tree(self, obj, tag, ctx):
        # lookup source for obj
        block_index = ctx.find_available_block_index(
            lambda: np.ndarray(len(obj.payload), dtype="uint8", buffer=obj.payload),
        )
        return {
            "block_index": block_index,
        }

    def from_yaml_tree(self, node, tag, ctx):
        block_index = node["block_index"]
        data = ctx.get_block_data_callback(block_index)()
        obj = BlockData(data.tobytes())
        return obj


class BlockExtension(Extension):
    tags = ["asdf://somewhere.org/tags/block_data-1.0.0"]
    converters = [BlockConverter()]
    extension_uri = "asdf://somewhere.org/extensions/block_data-1.0.0"


@contextlib.contextmanager
def with_extension(ext_class):
    with asdf.config_context() as cfg:
        cfg.add_extension(ext_class())
        yield


@with_extension(BlockExtension)
def test_roundtrip_block_data():
    a = BlockData(b"abcdefg")
    b = helpers.roundtrip_object(a)
    assert a.payload == b.payload


@with_extension(BlockExtension)
def test_block_converter_block_allocation(tmp_path):
    a = BlockData(b"abcdefg")

    # make a tree without the BlockData instance to avoid
    # the initial validate which will trigger block allocation
    af = asdf.AsdfFile({"a": None})
    # now assign to the tree item (avoiding validation)
    af["a"] = a

    # they currently hold storage settings
    fn = tmp_path / "test.asdf"
    af.write_to(fn)

    # if we read a file
    with asdf.open(fn, mode="rw") as af:
        fn2 = tmp_path / "test2.asdf"
        # there should be 1 block
        assert len(af._blocks.blocks) == 1
        # validate should use that block
        af.validate()
        assert len(af._blocks.blocks) == 1
        # as should write_to
        af.write_to(fn2)
        assert len(af._blocks.blocks) == 1
        # and update
        af.update()
        assert len(af._blocks.blocks) == 1


class BlockDataCallback:
    """An example object that uses the data callback to read block data"""

    def __init__(self, callback):
        self.callback = callback

    @property
    def data(self):
        return self.callback()


class BlockDataCallbackConverter(Converter):
    tags = ["asdf://somewhere.org/tags/block_data_callback-1.0.0"]
    types = [BlockDataCallback]

    def to_yaml_tree(self, obj, tag, ctx):
        block_index = ctx.find_available_block_index(obj.callback)
        return {
            "block_index": block_index,
        }

    def from_yaml_tree(self, node, tag, ctx):
        block_index = node["block_index"]

        obj = BlockDataCallback(ctx.get_block_data_callback(block_index))
        return obj


class BlockDataCallbackExtension(Extension):
    tags = ["asdf://somewhere.org/tags/block_data_callback-1.0.0"]
    converters = [BlockDataCallbackConverter()]
    extension_uri = "asdf://somewhere.org/extensions/block_data_callback-1.0.0"


@with_extension(BlockDataCallbackExtension)
def test_block_data_callback_converter(tmp_path):
    # use a callback that every time generates a new array
    # this would cause issues for the old block management as the
    # id(arr) would change every time
    a = BlockDataCallback(lambda: np.zeros(3, dtype="uint8"))

    tfn = tmp_path / "tmp.asdf"
    asdf.AsdfFile({"obj": a}).write_to(tfn)
    with asdf.open(tfn) as af:
        assert_array_equal(a.data, af["obj"].data)

    # make a tree without the BlockData instance to avoid
    # the initial validate which will trigger block allocation
    af = asdf.AsdfFile({"a": None})
    # now assign to the tree item (avoiding validation)
    af["a"] = a
    # write_to will use the block
    fn1 = tmp_path / "test.asdf"
    af.write_to(fn1)

    # if we read a file
    with asdf.open(fn1, mode="rw") as af:
        fn2 = tmp_path / "test2.asdf"
        # there should be 1 block
        assert len(af._blocks.blocks) == 1
        # validate should use that block
        af.validate()
        assert len(af._blocks.blocks) == 1
        # as should write_to
        af.write_to(fn2)
        assert len(af._blocks.blocks) == 1
        # and update
        af.update()
        assert len(af._blocks.blocks) == 1

    # check that data was preserved
    for fn in (fn1, fn2):
        with asdf.open(fn) as af:
            assert_array_equal(af["a"].data, a.data)


@with_extension(BlockDataCallbackExtension)
def test_block_with_callback_removal(tmp_path):
    fn1 = tmp_path / "test1.asdf"
    fn2 = tmp_path / "test2.asdf"

    a = BlockDataCallback(lambda: np.zeros(3, dtype="uint8"))
    b = BlockDataCallback(lambda: np.ones(3, dtype="uint8"))
    base_af = asdf.AsdfFile({"a": a, "b": b})
    base_af.write_to(fn1)

    for remove_key, check_key in [("a", "b"), ("b", "a")]:
        # check that removing one does not interfere with the other
        with asdf.open(fn1) as af:
            af[remove_key] = None
            af.write_to(fn2)
        with asdf.open(fn2) as af:
            af[check_key] = b.data
        # also test update
        # first copy fn1 to fn2
        with asdf.open(fn1) as af:
            af.write_to(fn2)
        with asdf.open(fn2, mode="rw") as af:
            af[remove_key] = None
            af.update()
            af[check_key] = b.data


class MultiBlockData:
    def __init__(self, data):
        self.data = data
        self.keys = []


class MultiBlockConverter(Converter):
    tags = ["asdf://somewhere.org/tags/multi_block_data-1.0.0"]
    types = [MultiBlockData]

    def to_yaml_tree(self, obj, tag, ctx):
        if not len(obj.keys):
            obj.keys = [ctx.generate_block_key() for _ in obj.data]
        indices = [ctx.find_available_block_index(d, k) for d, k in zip(obj.data, obj.keys)]
        return {
            "indices": indices,
        }

    def from_yaml_tree(self, node, tag, ctx):
        indices = node["indices"]
        keys = [ctx.generate_block_key() for _ in indices]
        cbs = [ctx.get_block_data_callback(i, k) for i, k in zip(indices, keys)]
        obj = MultiBlockData([cb() for cb in cbs])
        obj.keys = keys
        return obj


class MultiBlockExtension(Extension):
    tags = ["asdf://somewhere.org/tags/multi_block_data-1.0.0"]
    converters = [MultiBlockConverter()]
    extension_uri = "asdf://somewhere.org/extensions/multi_block_data-1.0.0"


@with_extension(MultiBlockExtension)
def test_mutli_block():
    a = MultiBlockData([np.arange(3, dtype="uint8") for i in range(3)])
    b = helpers.roundtrip_object(a)
    assert len(a.data) == len(b.data)
    assert [np.testing.assert_array_equal(aa, ab) for aa, ab in zip(a.data, b.data)]


class SharedBlockData:
    def __init__(self, callback):
        self.callback = callback

    @property
    def data(self):
        return self.callback()


class SharedBlockConverter(Converter):
    tags = ["asdf://somewhere.org/tags/shared_block_data-1.0.0"]
    types = [SharedBlockData]
    _return_invalid_keys = False

    def to_yaml_tree(self, obj, tag, ctx):
        # lookup source for obj
        block_index = ctx.find_available_block_index(
            lambda: obj.data,
        )
        return {
            "block_index": block_index,
        }

    def from_yaml_tree(self, node, tag, ctx):
        block_index = node["block_index"]
        callback = ctx.get_block_data_callback(block_index)
        obj = SharedBlockData(callback)
        return obj


class SharedBlockExtension(Extension):
    tags = ["asdf://somewhere.org/tags/shared_block_data-1.0.0"]
    converters = [SharedBlockConverter()]
    extension_uri = "asdf://somewhere.org/extensions/shared_block_data-1.0.0"


@with_extension(SharedBlockExtension)
def test_shared_block_reassignment(tmp_path):
    fn = tmp_path / "test.asdf"
    arr1 = np.arange(10, dtype="uint8")
    arr2 = np.arange(5, dtype="uint8")
    a = SharedBlockData(lambda: arr1)
    b = SharedBlockData(lambda: arr1)
    asdf.AsdfFile({"a": a, "b": b}).write_to(fn)
    with asdf.open(fn, mode="rw") as af:
        af["b"].callback = lambda: arr2
        af.update()
    with asdf.open(fn) as af:
        np.testing.assert_array_equal(af["a"].data, arr1)
        np.testing.assert_array_equal(af["b"].data, arr2)


@with_extension(SharedBlockExtension)
def test_shared_block_obj_removal(tmp_path):
    fn = tmp_path / "test.asdf"
    arr1 = np.arange(10, dtype="uint8")
    a = SharedBlockData(lambda: arr1)
    b = SharedBlockData(lambda: arr1)
    asdf.AsdfFile({"a": a, "b": b}).write_to(fn)
    with asdf.open(fn, mode="rw") as af:
        af["b"] = None
        del b
        gc.collect(2)
        af.update()
    with asdf.open(fn) as af:
        np.testing.assert_array_equal(af["a"].data, arr1)
        assert af["b"] is None



import numpy
import pytest

import asdf


@pytest.mark.parametrize("lazy_load", [True, False])
@pytest.mark.parametrize("include_block_index", [True, False])
@pytest.mark.parametrize("index", [0, 1, 2])
def test_block_checksums_only_checked_for_first_block_if_index_exists(tmp_path, index, include_block_index, lazy_load):
    """
    Block checksums are only checked for first block if a block index is present

    https://github.com/asdf-format/asdf/issues/1541
    """
    fn = tmp_path / "test.asdf"
    arrs = [numpy.zeros(1) + i for i in range(3)]
    asdf.AsdfFile({"arrs": arrs}).write_to(fn, include_block_index=include_block_index)

    # read file to get block offset
    with asdf.open(fn, lazy_load=False) as af:
        checksum_offset = af._blocks.blocks[index].offset + 2 + 4 + 4 + 8 + 8 + 8

    # now modify the block checksum
    with open(fn, "r+b") as f:
        f.seek(checksum_offset)
        v = f.read(1)[0]
        f.seek(checksum_offset)
        f.write(bytes([v + 1]))

    # and check that it raises an error
    with pytest.raises(ValueError, match=r".* does not match given checksum"):
        with asdf.open(fn, lazy_load=lazy_load, validate_checksums=True) as af:
            sum([a[0] for a in af["arrs"]])



import copy

import pytest

from asdf._block.options import Options
from asdf.config import config_context

valid_storage_types = ["internal", "external", "streamed", "inline"]
valid_default_storage_types = [st for st in valid_storage_types if st != "streamed"]
valid_compression_types = [None, "zlib", "bzp2", "lz4", ""]

invalid_storage_types = ["foo", "bar"]
invalid_compression_types = ["input", "foo"]


@pytest.mark.parametrize("storage", valid_storage_types)
def test_set_storage_init(storage):
    o = Options(storage)
    assert o.storage_type == storage


@pytest.mark.parametrize("storage", valid_default_storage_types)
def test_default_storage_init(storage):
    with config_context() as cfg:
        cfg.all_array_storage = storage
        o = Options()
        assert o.storage_type == storage


@pytest.mark.parametrize("storage", valid_storage_types)
def test_set_storage_attr(storage):
    # start with a different storage type
    o = Options("internal" if storage == "external" else "external")
    o.storage_type = storage
    assert o.storage_type == storage


@pytest.mark.parametrize("compression", valid_compression_types)
def test_set_compression_attr(compression):
    o = Options("internal")
    o.compression = compression
    # allow "" to become None, both are falsey
    assert o.compression == compression if compression else not o.compression


@pytest.mark.parametrize("compression", valid_compression_types)
def test_set_compression_init(compression):
    o = Options("internal", compression)
    # allow "" to become None, both are falsey
    assert o.compression == compression if compression else not o.compression


def test_set_compression_kwargs_attr():
    o = Options("internal")
    o.compression_kwargs = {"foo": 1}
    assert o.compression_kwargs == {"foo": 1}


def test_set_compression_kwargs_init():
    o = Options("internal", compression_kwargs={"foo": 1})
    assert o.compression_kwargs == {"foo": 1}


def test_default_compression():
    o = Options("internal")
    assert o.compression is None


@pytest.mark.parametrize("invalid_storage", invalid_storage_types)
def test_invalid_storage_type_init(invalid_storage):
    with pytest.raises(ValueError, match=r"array_storage must be one of.*"):
        Options(invalid_storage)


@pytest.mark.parametrize("invalid_storage", invalid_storage_types)
def test_invalid_storage_attr(invalid_storage):
    o = Options("internal")
    with pytest.raises(ValueError, match=r"array_storage must be one of.*"):
        o.storage_type = invalid_storage


@pytest.mark.parametrize("invalid_compression", invalid_compression_types)
def test_invalid_compression_attr(invalid_compression):
    o = Options("internal")
    with pytest.raises(ValueError, match=r"Invalid compression.*"):
        o.compression = invalid_compression


@pytest.mark.parametrize("invalid_compression", invalid_compression_types)
def test_invalid_compression_init(invalid_compression):
    with pytest.raises(ValueError, match=r"Invalid compression.*"):
        Options("internal", invalid_compression)


@pytest.mark.parametrize("storage", valid_storage_types)
@pytest.mark.parametrize("compression", valid_compression_types)
@pytest.mark.parametrize("compression_kwargs", [None, {"foo": 1}])
def test_copy(storage, compression, compression_kwargs):
    o = Options(storage, compression, compression_kwargs)
    o2 = copy.copy(o)
    assert o2 is not o
    assert o2.storage_type == storage
    # allow "" to become None, both are falsey
    assert o2.compression == compression if compression else not o2.compression
    # allow None to become {}, both are falsey
    assert o2.compression_kwargs == compression_kwargs if compression_kwargs else not o2.compression_kwargs



import yaml

from ._extension import Extension
from ._tag import TagDefinition


class ManifestExtension(Extension):
    """
    Extension implementation that reads the extension URI, ASDF
    Standard requirement, and tag list from a manifest document.

    Parameters
    ----------
    manifest : dict
        Parsed manifest.
    converters : iterable of asdf.extension.Converter, optional
        Converter instances for the tags and Python types
        supported by this extension.
    compressors : iterable of asdf.extension.Compressor, optional
        Compressor instances to support additional binary
        block compression options.
    validators : iterable of asdf.extension.Validator, optional
        Validator instances to support validation of custom
        schema properties.
    legacy_class_names : iterable of str, optional
        Fully-qualified class names used by older versions
        of this extension.
    """

    @classmethod
    def from_uri(cls, manifest_uri, **kwargs):
        """
        Construct the extension using the manifest with the
        specified URI.  The manifest document must be registered
        with ASDF's resource manager.

        Parameters
        ----------
        manifest_uri : str
            Manifest URI.

        See the class docstring for details on keyword parameters.
        """
        from asdf.config import get_config

        manifest = yaml.safe_load(get_config().resource_manager[manifest_uri])
        return cls(manifest, **kwargs)

    def __init__(self, manifest, *, legacy_class_names=None, converters=None, compressors=None, validators=None):
        self._manifest = manifest

        if legacy_class_names is None:
            self._legacy_class_names = []
        else:
            self._legacy_class_names = legacy_class_names

        if converters is None:
            self._converters = []
        else:
            self._converters = converters

        if compressors is None:
            self._compressors = []
        else:
            self._compressors = compressors

        if validators is None:
            self._validators = []
        else:
            self._validators = validators

    @property
    def extension_uri(self):
        return self._manifest["extension_uri"]

    @property
    def legacy_class_names(self):
        return self._legacy_class_names

    @property
    def asdf_standard_requirement(self):
        version = self._manifest.get("asdf_standard_requirement", None)
        if version is None:
            return None

        if isinstance(version, str):
            return f"=={version}"

        specifiers = []
        for prop, operator in [("gt", ">"), ("gte", ">="), ("lt", "<"), ("lte", "<=")]:
            value = version.get(prop)
            if value:
                specifiers.append(f"{operator}{value}")
        return ",".join(specifiers)

    @property
    def converters(self):
        return self._converters

    @property
    def compressors(self):
        return self._compressors

    @property
    def validators(self):
        return self._validators

    @property
    def tags(self):
        result = []
        for tag in self._manifest.get("tags", []):
            if isinstance(tag, str):
                # ExtensionProxy knows how to handle str tags.
                result.append(tag)
            elif isinstance(tag, dict):
                result.append(
                    TagDefinition(
                        tag["tag_uri"],
                        schema_uris=tag.get("schema_uri"),
                        title=tag.get("title"),
                        description=tag.get("description"),
                    ),
                )
            else:
                msg = "Malformed manifest document"
                raise TypeError(msg)
        return result



import re

import pytest

import asdf
from asdf.testing import helpers


def make_complex_asdf(string):
    yaml = f"""
a: !core/complex-1.0.0
  {string}
    """

    return helpers.yaml_to_asdf(yaml)


@pytest.mark.parametrize(
    "invalid",
    [
        "3 + 4i",
        "3+-4i",
        "3-+4i",
        "3i+4i",
        "X3+4iX",
        "3+X4i",
        "3+4",
        "3i+43+4z",
        "3.+4i",
        "3+4.i",
        "3e-4.0+4i",
        "3+4e4.0i",
        "",
    ],
)
def test_invalid_complex(invalid):
    with pytest.raises(asdf.ValidationError), asdf.open(make_complex_asdf(invalid)):
        pass


@pytest.mark.parametrize(
    "valid",
    [
        "3+4j",
        "(3+4j)",
        ".3+4j",
        "3+.4j",
        "3e10+4j",
        "3e-10+4j",
        "3+4e10j",
        "3.0+4j",
        "3+4.0j",
        "3.0+4.0j",
        "3+4e-10j",
        "3+4J",
        "3+4i",
        "3+4I",
        "inf",
        "inf+infj",
        "inf+infi",
        "infj",
        "infi",
        "INFi",
        "INFI",
        "3+infj",
        "inf+4j",
    ],
)
def test_valid_complex(valid):
    with asdf.open(make_complex_asdf(valid)) as af:
        assert af.tree["a"] == complex(re.sub(r"[iI]$", r"j", valid))


@pytest.mark.parametrize(
    "valid",
    ["nan", "nan+nanj", "nan+nani", "nanj", "nani", "NANi", "NANI", "3+nanj", "nan+4j"],
)
def test_valid_nan_complex(valid):
    with asdf.open(make_complex_asdf(valid)):
        pass


def test_roundtrip(tmp_path):
    values = {
        "a": 0 + 0j,
        "b": 1 + 1j,
        "c": -1 + 1j,
        "d": -1 - 1j,
    }

    fn = tmp_path / "test.asdf"
    asdf.AsdfFile({"values": values}).write_to(fn)
    with asdf.open(fn) as af:
        result = af["values"]
        assert len(values) == len(result)
        for key, value in values.items():
            assert result[key] == value



import numpy as np
import pytest

import asdf

# botocore uses the deprecated utcnow
pytestmark = pytest.mark.filterwarnings(r"ignore:datetime\.datetime\.utcnow\(\) is deprecated.*:DeprecationWarning")


@pytest.fixture()
def bucket_url(s3):
    s3fs = pytest.importorskip("s3fs")
    bucket_name = "test"
    s3.create_bucket(Bucket="test")
    base_url = f"s3://{bucket_name}"

    fs = s3fs.S3FileSystem()

    arr = np.arange(42)
    af = asdf.AsdfFile({"arr": arr})

    url = f"{base_url}/test.asdf"
    with fs.open(url, mode="wb") as f:
        af.write_to(f)
    yield url


def test_s3fs(bucket_url):
    fsspec = pytest.importorskip("fsspec")

    with fsspec.open(bucket_url, mode="rb") as f:
        with asdf.open(f, memmap=False) as af:
            assert af["arr"][-1] == 41


def test_native(bucket_url):
    with asdf.open(bucket_url, "r") as af:
        assert af["arr"][-1] == 41



import sys
from functools import lru_cache

from asdf.tagged import Tagged
from asdf.util import get_class_name, uri_match

from ._extension import ExtensionProxy


def _resolve_type(path):
    """
    Convert a class path (like the string "asdf.AsdfFile") to a
    class (``asdf.AsdfFile``) only if the module implementing the
    class has already been imported.

    Parameters
    ----------

    path : str
        Path/name of class (for example, "asdf.AsdfFile")

    Returns
    -------

    typ : class or None
        The class (if it's already been imported) or None
    """
    if "." not in path:
        # check if this path is a module
        if path in sys.modules:
            return sys.modules[path]
        return None
    # this type is part of a module
    module_name, type_name = path.rsplit(".", maxsplit=1)
    # if the module is not imported, don't index it
    if module_name not in sys.modules:
        return None
    module = sys.modules[module_name]
    if not hasattr(module, type_name):
        # the imported module does not have this class, perhaps
        # it is dynamically created so do not index it yet
        return None
    return getattr(module, type_name)


class ExtensionManager:
    """
    Wraps a list of extensions and indexes their converters
    by tag and by Python type.

    Parameters
    ----------
    extensions : iterable of asdf.extension.Extension
        List of enabled extensions to manage.  Extensions placed earlier
        in the list take precedence.
    """

    def __init__(self, extensions):
        self._extensions = [ExtensionProxy.maybe_wrap(e) for e in extensions]

        self._tag_defs_by_tag = {}
        self._converters_by_tag = {}

        # To optimize performance converters can be registered using either:
        # - the class/type they convert
        # - the name/path (string) of the class/type they convert
        # This allows the registration to continue without importing
        # every module for every extension (which would be needed to turn
        # the class paths into proper classes). Using class paths can be
        # complicated by packages that have private implementations of
        # classes that are exposed at a different 'public' location.
        # These private classes may change between minor versions
        # and would break converters that are registered using the private
        # class path. However, often libraries do not modify the module
        # of the 'public' class (so inspecting the class path returns
        # the private class path). One example of this in asdf is
        # Converter (exposed as ``asdf.extension.Converter`` but with
        # a class path of ``asdf.extension._converter.Converter``).
        # To allow converters to be registered with the public location
        # we will need to attempt to import the public class path
        # and then register the private class path after the class is
        # imported. We don't want to do this unnecessarily and since
        # class instances do not contain the public class path
        # we adopt a strategy of checking class paths and only
        # registering those that have already been imported. This
        # is ok because asdf will only use the converter type
        # when attempting to serialize an object in memory (so the
        # public class path will already be imported at the time
        # the converter is needed).

        # first we store the converters in the order they are discovered
        # the key here can either be a class path (str) or class (type)
        converters_by_type = {}
        validators = set()

        for extension in self._extensions:
            for tag_def in extension.tags:
                if tag_def.tag_uri not in self._tag_defs_by_tag:
                    self._tag_defs_by_tag[tag_def.tag_uri] = tag_def
            for converter in extension.converters:
                for tag in converter.tags:
                    if tag not in self._converters_by_tag:
                        self._converters_by_tag[tag] = converter
                for typ in converter.types:
                    if typ not in converters_by_type:
                        converters_by_type[typ] = converter

            validators.update(extension.validators)

        self._converters_by_class_path = {}
        self._converters_by_type = {}

        for type_or_path, converter in converters_by_type.items():
            if isinstance(type_or_path, str):
                path = type_or_path
                typ = _resolve_type(path)
                if typ is None:
                    if path not in self._converters_by_class_path:
                        self._converters_by_class_path[path] = converter
                        continue
            else:
                typ = type_or_path
            if typ not in self._converters_by_type:
                self._converters_by_type[typ] = converter

        self._validator_manager = _get_cached_validator_manager(tuple(validators))

    @property
    def extensions(self):
        """
        Get the list of extensions.

        Returns
        -------
        list of asdf.extension.ExtensionProxy
        """
        return self._extensions

    def handles_tag(self, tag):
        """
        Return `True` if the specified tag is handled by a
        converter.

        Parameters
        ----------
        tag : str
            Tag URI.

        Returns
        -------
        bool
        """
        return tag in self._converters_by_tag

    def handles_type(self, typ):
        """
        Returns `True` if the specified Python type is handled
        by a converter.

        Parameters
        ----------
        typ : type

        Returns
        -------
        bool
        """
        if typ in self._converters_by_type:
            return True
        self._index_converters()
        return typ in self._converters_by_type

    def handles_tag_definition(self, tag):
        """
        Return `True` if the specified tag has a definition.

        Parameters
        ----------
        tag : str
            Tag URI.

        Returns
        -------
        bool
        """
        return tag in self._tag_defs_by_tag

    def get_tag_definition(self, tag):
        """
        Get the tag definition for the specified tag.

        Parameters
        ----------
        tag : str
            Tag URI.

        Returns
        -------
        asdf.extension.TagDefinition

        Raises
        ------
        KeyError
            Unrecognized tag URI.
        """
        try:
            return self._tag_defs_by_tag[tag]
        except KeyError:
            msg = f"No support available for YAML tag '{tag}'.  You may need to install a missing extension."
            raise KeyError(msg) from None

    def get_converter_for_tag(self, tag):
        """
        Get the converter for the specified tag.

        Parameters
        ----------
        tag : str
            Tag URI.

        Returns
        -------
        asdf.extension.Converter

        Raises
        ------
        KeyError
            Unrecognized tag URI.
        """
        try:
            return self._converters_by_tag[tag]
        except KeyError:
            msg = f"No support available for YAML tag '{tag}'.  You may need to install a missing extension."
            raise KeyError(msg) from None

    def get_converter_for_type(self, typ):
        """
        Get the converter for the specified Python type.

        Parameters
        ----------
        typ : type

        Returns
        -------
        asdf.extension.Converter

        Raises
        ------
        KeyError
            Unrecognized type.
        """
        if typ not in self._converters_by_type:
            self._index_converters()
        try:
            return self._converters_by_type[typ]
        except KeyError:
            msg = (
                f"No support available for Python type '{get_class_name(typ, instance=False)}'.  "
                "You may need to install or enable an extension."
            )
            raise KeyError(msg) from None

    def _index_converters(self):
        """
        Search _converters_by_class_path for paths (strings) that
        refer to classes that are currently imported. For imported
        classes, add them to _converters_by_class (if the class
        doesn't already have a converter).
        """
        # search class paths to find ones that are imported
        for class_path in list(self._converters_by_class_path):
            typ = _resolve_type(class_path)
            if typ is None:
                continue
            if typ not in self._converters_by_type:
                self._converters_by_type[typ] = self._converters_by_class_path[class_path]
            del self._converters_by_class_path[class_path]

    @property
    def validator_manager(self):
        return self._validator_manager


def get_cached_extension_manager(extensions):
    """
    Get a previously created ExtensionManager for the specified
    extensions, or create and cache one if necessary.  Building
    the manager is expensive, so it helps performance to reuse
    it when possible.

    Parameters
    ----------
    extensions : list of asdf.extension.Extension

    Returns
    -------
    asdf.extension.ExtensionManager
    """
    from ._extension import ExtensionProxy

    # The tuple makes the extensions hashable so that we
    # can pass them to the lru_cache method.  The ExtensionProxy
    # overrides __hash__ to return the hashed object id of the wrapped
    # extension, so this will method will only return the same
    # ExtensionManager if the list contains identical extension
    # instances in identical order.
    extensions = tuple(ExtensionProxy.maybe_wrap(e) for e in extensions)

    return _get_cached_extension_manager(extensions)


@lru_cache
def _get_cached_extension_manager(extensions):
    return ExtensionManager(extensions)


class ValidatorManager:
    """
    Wraps a list of custom validators and indexes them by schema property.

    Parameters
    ----------
    validators : iterable of asdf.extension.Validator
        List of validators to manage.
    """

    def __init__(self, validators):
        self._validators = list(validators)

        self._validators_by_schema_property = {}
        for validator in self._validators:
            if validator.schema_property not in self._validators_by_schema_property:
                self._validators_by_schema_property[validator.schema_property] = set()
            self._validators_by_schema_property[validator.schema_property].add(validator)

        self._jsonschema_validators_by_schema_property = {}
        for schema_property in self._validators_by_schema_property:
            self._jsonschema_validators_by_schema_property[schema_property] = self._get_jsonschema_validator(
                schema_property,
            )

    def validate(self, schema_property, schema_property_value, node, schema):
        """
        Validate an ASDF tree node against custom validators for a schema property.

        Parameters
        ----------
        schema_property : str
            Name of the schema property (identifies the validator(s) to use).
        schema_property_value : object
            Value of the schema property.
        node : asdf.tagged.Tagged
            The ASDF node to validate.
        schema : dict
            The schema object that contains the property that triggered
            the validation.

        Yields
        ------
        asdf.exceptions.ValidationError
        """
        if schema_property in self._validators_by_schema_property:
            for validator in self._validators_by_schema_property[schema_property]:
                if _validator_matches(validator, node):
                    yield from validator.validate(schema_property_value, node, schema)

    def get_jsonschema_validators(self):
        """
        Get a dictionary of validator methods suitable for use
        with the jsonschema library.

        Returns
        -------
        dict of str: callable
        """
        return dict(self._jsonschema_validators_by_schema_property)

    def _get_jsonschema_validator(self, schema_property):
        def _validator(_, schema_property_value, node, schema):
            return self.validate(schema_property, schema_property_value, node, schema)

        return _validator


def _validator_matches(validator, node):
    if any(t == "**" for t in validator.tags):
        return True

    if not isinstance(node, Tagged):
        return False

    return any(uri_match(t, node._tag) for t in validator.tags)


@lru_cache
def _get_cached_validator_manager(validators):
    return ValidatorManager(validators)



import datetime

import asdf
from asdf.tags.core import AsdfObject, ExtensionMetadata, HistoryEntry, Software, SubclassMetadata
from asdf.testing import helpers


def test_asdf_object():
    asdf_object = AsdfObject({"foo": "bar"})

    result = helpers.roundtrip_object(asdf_object)

    assert result == asdf_object


def test_extension_metadata():
    metadata = ExtensionMetadata(
        extension_class="foo.extension.FooExtension",
        extension_uri="http://foo.biz/extensions/foo-1.0.0",
        software=Software(name="FooSoft", version="1.5"),
    )

    result = helpers.roundtrip_object(metadata)

    assert result == metadata


def test_extension_metadata_extra_properties():
    yaml = """
metadata: !core/extension_metadata-1.0.0
  extension_class: foo.extension.FooExtension
  software: !core/software-1.0.0
    name: FooSoft
    version: "1.5"
  extension_uri: http://foo.biz/extensions/foo-1.0.0
  extra: property
    """

    buff = helpers.yaml_to_asdf(yaml)

    with asdf.open(buff) as af:
        assert af["metadata"].extension_class == "foo.extension.FooExtension"
        assert af["metadata"].software["name"] == "FooSoft"
        assert af["metadata"].software["version"] == "1.5"
        assert af["metadata"].extension_uri == "http://foo.biz/extensions/foo-1.0.0"
        assert af["metadata"]["extra"] == "property"


def test_software():
    software = Software(
        name="FooSoft",
        version="1.5.0",
        author="The Foo Developers",
        homepage="http://nowhere.org",
        extra="property",
    )

    result = helpers.roundtrip_object(software)

    assert result == software


def test_history_entry(tmp_path):
    history_entry = HistoryEntry(
        description="Some history happened here",
        time=datetime.datetime.now(),
        software=[Software(name="FooSoft", version="1.5.0")],
        extra="property",
    )

    fn = tmp_path / "test.asdf"
    asdf.AsdfFile({"obj": history_entry}).write_to(fn)
    with asdf.open(fn) as af:
        assert af["obj"] == history_entry


def test_subclass_metadata():
    subclass_metadata = SubclassMetadata(name="SomeCoolSubclass")

    # support for this was dropped in 1.6.0
    result = helpers.roundtrip_object(subclass_metadata, version="1.5.0")

    assert result == subclass_metadata



import numpy as np
import pytest

import asdf


@pytest.mark.parametrize("memmap", [True, False])
def test_external_blocks_always_lazy_loaded_and_memmapped(tmp_path, memmap):
    """
    External blocks are always lazy loaded and memmapped

    https://github.com/asdf-format/asdf/issues/1525
    """

    fn = tmp_path / "test.asdf"
    arr = np.arange(10)
    af = asdf.AsdfFile({"arr": arr})
    af.set_array_storage(arr, "external")
    af.write_to(fn)

    with asdf.open(fn, memmap=memmap) as af:
        # check that block is external
        source = af["arr"]._source
        assert isinstance(source, str)

        # check if block is memmapped
        if not memmap:
            assert not isinstance(af["arr"].base, np.memmap)
        else:
            assert isinstance(af["arr"].base, np.memmap)



"""
Support for Converter, the new API for serializing custom
types.  Will eventually replace the `asdf.types` module.
"""

import abc

from asdf.util import get_class_name, uri_match


class Converter(abc.ABC):
    """
    Abstract base class for plugins that convert nodes from the
    parsed YAML tree into custom objects, and vice versa.

    Implementing classes must provide the `tags` and `types`
    properties and `to_yaml_tree` and `from_yaml_tree` methods.
    The ``select_tag`` method is optional.

    If implemented, ``select_tag`` should accept 3 parameters

        obj : object
            Instance of the custom type being converted.  Guaranteed
            to be an instance of one of the types listed in the
            `types` property.
        tags : list of str
            List of active tags to choose from.  Guaranteed to match
            one of the tag patterns listed in the 'tags' property.
        ctx : asdf.extension.SerializationContext
            Context of the current serialization request.

    and return a str, the selected tag (should be one of tags) or
    `None` which will trigger the result of ``to_yaml_tree`` to be
    used to look up the next converter for this object.

    The ``lazy`` attribute is optional. If ``True`` asdf will
    pass "lazy" objects to the converter. If ``False`` (or not
    defined) asdf will convert all child objects before calling
    `from_yaml_tree`.

    The ``to_info`` method is optional. If implemented it must
    accept 1 parameter ``obj` which is a tree node/custom
    object and return a container (list, tuple, dict) containing
    information about that object to display during ``AsdfFile.info``.
    """

    @classmethod
    def __subclasshook__(cls, class_):
        if cls is Converter:
            return (
                hasattr(class_, "tags")
                and hasattr(class_, "types")
                and hasattr(class_, "to_yaml_tree")
                and hasattr(class_, "from_yaml_tree")
            )
        return NotImplemented  # pragma: no cover

    @property
    @abc.abstractmethod
    def tags(self):
        """
        Get the YAML tags that this converter is capable of
        handling.  URI patterns are permitted, see
        `asdf.util.uri_match` for details.

        Returns
        -------
        iterable of str
            Tag URIs or URI patterns.
        """

    @property
    @abc.abstractmethod
    def types(self):
        """
        Get the Python types that this converter is capable of
        handling.

        Returns
        -------
        iterable of str or type
            If str, the fully qualified class name of the type.
        """

    @abc.abstractmethod
    def to_yaml_tree(self, obj, tag, ctx):
        """
        Convert an object into a node suitable for YAML serialization.
        This method is not responsible for writing actual YAML; rather, it
        converts an instance of a custom type to a built-in Python object type
        (such as dict, list, str, or number), which can then be automatically
        serialized to YAML as needed.

        For container types returned by this method (dict or list),
        the children of the container need not themselves be converted.
        Any list elements or dict values will be converted by subsequent
        calls to to_yaml_tree implementations.

        The returned node must be an instance of `dict`, `list`, or `str`.
        Children may be any type supported by an available Converter.

        Parameters
        ----------
        obj : object
            Instance of a custom type to be serialized.  Guaranteed to
            be an instance of one of the types listed in the `types`
            property.
        tag : str
            The tag identifying the YAML type that ``obj`` should be
            converted into.  Selected by a call to this converter's
            select_tag method.
        ctx : asdf.extension.SerializationContext
            The context of the current serialization request.

        Returns
        -------
        dict or list or str
            The YAML node representation of the object.
        """

    @abc.abstractmethod
    def from_yaml_tree(self, node, tag, ctx):
        """
        Convert a YAML node into an instance of a custom type.

        For container types received by this method (dict or list),
        the children of the container will have already been converted
        by prior calls to from_yaml_tree implementations unless
        ``lazy_tree`` was set to ``True`` for `asdf.open`. With a lazy
        tree the container types will be `asdf.lazy_nodes` (which act
        like dict or list but convert child objects when accessed).

        Note on circular references: trees that reference themselves
        among their descendants must be handled with care.  Most
        implementations need not concern themselves with this case, but
        if the custom type supports circular references, then the
        implementation of this method will need to return a generator.
        Consult the documentation for more details.

        Parameters
        ----------
        node : dict or list or str
            The YAML node to convert.
        tag : str
            The YAML tag of the object being converted.
        ctx : asdf.extension.SerializationContext
            The context of the current deserialization request.

        Returns
        -------
        object
            An instance of one of the types listed in the `types` property,
            or a generator that yields such an instance.
        """


class ConverterProxy(Converter):
    """
    Proxy that wraps a `Converter` and provides default
    implementations of optional methods.
    """

    def __init__(self, delegate, extension):
        if not isinstance(delegate, Converter):
            msg = "Converter must implement the asdf.extension.Converter interface"
            raise TypeError(msg)

        self._delegate = delegate
        self._extension = extension
        self._class_name = get_class_name(delegate)

        # Sort these out up-front so that errors are raised when the extension is loaded
        # and not in the middle of the user's session.  The extension will fail to load
        # and a warning will be emitted, but it won't crash the program.

        relevant_tags = set()
        for tag in delegate.tags:
            if isinstance(tag, str):
                relevant_tags.update(t.tag_uri for t in extension.tags if uri_match(tag, t.tag_uri))
            else:
                msg = "Converter property 'tags' must contain str values"
                raise TypeError(msg)

        if len(relevant_tags) > 1 and not hasattr(delegate, "select_tag"):
            msg = "Converter handles multiple tags for this extension, but does not implement a select_tag method."
            raise RuntimeError(msg)

        self._tags = sorted(relevant_tags)

        self._types = []

        if not len(self._tags) and not hasattr(delegate, "select_tag"):
            # this converter supports no tags so don't inspect the types
            return

        for typ in delegate.types:
            if isinstance(typ, (str, type)):
                self._types.append(typ)
            else:
                msg = "Converter property 'types' must contain str or type values"
                raise TypeError(msg)

    @property
    def lazy(self):
        """
        Boolean indicating if this Converter supports "lazy" node objects

        Returns
        -------
        bool
        """
        return getattr(self._delegate, "lazy", False)

    @property
    def tags(self):
        """
        Get the list of tag URIs that this converter is capable of
        handling.

        Returns
        -------
        list of str
        """
        return self._tags

    @property
    def types(self):
        """
        Get the Python types that this converter is capable of
        handling.

        Returns
        -------
        list of type or str
        """
        return self._types

    def select_tag(self, obj, ctx):
        """
        Select the tag to use when converting an object to YAML.

        Parameters
        ----------
        obj : object
            Instance of the custom type being converted.
        ctx : asdf.extension.SerializationContext
            Serialization parameters.

        Returns
        -------
        str or None
            Selected tag or `None` to defer conversion.
        """
        method = getattr(self._delegate, "select_tag", None)
        if method is None:
            return self._tags[0]

        return method(obj, self._tags, ctx)

    def to_yaml_tree(self, obj, tag, ctx):
        """
        Convert an object into a node suitable for YAML serialization.

        Parameters
        ----------
        obj : object
            Instance of a custom type to be serialized.
        tag : str
            The tag identifying the YAML type that ``obj`` should be
            converted into.
        ctx : asdf.extension.SerializationContext
            Serialization parameters.

        Returns
        -------
        object
            The YAML node representation of the object.
        """
        return self._delegate.to_yaml_tree(obj, tag, ctx)

    def from_yaml_tree(self, node, tag, ctx):
        """
        Convert a YAML node into an instance of a custom type.

        Parameters
        ----------
        tree : dict or list or str
            The YAML node to convert.
        tag : str
            The YAML tag of the object being converted.
        ctx : asdf.extension.SerializationContext
            Serialization parameters.

        Returns
        -------
        object
        """
        return self._delegate.from_yaml_tree(node, tag, ctx)

    def to_info(self, obj):
        """
        Convert an object to a container with items further
        defining information about this node. This method
        is used for "info" and not used for serialization.

        Parameters
        ----------
        obj : object
            Instance of a custom type to get "info" for.

        Returns
        -------
        object
            Must be a container (list, tuple, dict) with
            items providing "info" about ``obj``.
        """
        if not hasattr(self._delegate, "to_info"):
            return obj
        return self._delegate.to_info(obj)

    @property
    def delegate(self):
        """
        Get the wrapped converter instance.

        Returns
        -------
        asdf.extension.Converter
        """
        return self._delegate

    @property
    def extension(self):
        """
        Get the extension that provided this converter.

        Returns
        -------
        asdf.extension.ExtensionProxy
        """
        return self._extension

    @property
    def package_name(self):
        """
        Get the name of the Python package of this converter's
        extension.  This may not be the same package that implements
        the converter's class.

        Returns
        -------
        str or None
            Package name, or `None` if the extension was added at runtime.
        """
        return self.extension.package_name

    @property
    def package_version(self):
        """
        Get the version of the Python package of this converter's
        extension.  This may not be the same package that implements
        the converter's class.

        Returns
        -------
        str or None
            Package version, or `None` if the extension was added at runtime.
        """
        return self.extension.package_version

    @property
    def class_name(self):
        """
        Get the fully qualified class name of this converter.

        Returns
        -------
        str
        """
        return self._class_name

    def __eq__(self, other):
        if isinstance(other, ConverterProxy):
            return other.delegate is self.delegate and other.extension is self.extension

        return False

    def __hash__(self):
        return hash((id(self.delegate), id(self.extension)))

    def __repr__(self):
        package_description = "(none)" if self.package_name is None else f"{self.package_name}=={self.package_version}"

        return f"<ConverterProxy class: {self.class_name} package: {package_description}>"



import enum
import importlib.util
import inspect
import math
import re
import struct
import sys
from functools import lru_cache

import numpy as np
import yaml

from . import constants

# The standard library importlib.metadata returns duplicate entrypoints
# for all python versions up to and including 3.11
# https://github.com/python/importlib_metadata/issues/410#issuecomment-1304258228
# see PR https://github.com/asdf-format/asdf/pull/1260
# see issue https://github.com/asdf-format/asdf/issues/1254
if sys.version_info >= (3, 12):
    pass
else:
    pass


# We're importing our own copy of urllib.parse because
# we need to patch it to support asdf:// URIs, but it'd
# be irresponsible to do this for all users of a
# standard library.
urllib_parse_spec = importlib.util.find_spec("urllib.parse")
_patched_urllib_parse = importlib.util.module_from_spec(urllib_parse_spec)
urllib_parse_spec.loader.exec_module(_patched_urllib_parse)
del urllib_parse_spec

# urllib.parse needs to know that it should treat asdf://
# URIs like http:// URIs for the purposes of joining
# a relative path to a base URI.
_patched_urllib_parse.uses_relative.append("asdf")
_patched_urllib_parse.uses_netloc.append("asdf")


__all__ = [
    "FileType",
    "NotSet",
    "calculate_padding",
    "get_array_base",
    "get_base_uri",
    "get_class_name",
    "get_file_type",
    "load_yaml",
    "uri_match",
]


def load_yaml(init, tagged=False):
    """
    Load just the yaml portion of an ASDF file

    Parameters
    ----------

    init : filename or file-like
        If file-like this must be opened in binary mode.

    tagged: bool, optional
        Return tree with instances of `asdf.tagged.Tagged` this
        can be helpful if the yaml tags are of interest.
        If False, the tree will only contain basic python types
        (see the pyyaml ``BaseLoader`` documentation).

    Returns
    -------

    tree : dict
        Dictionary representing the ASDF tree
    """

    from .generic_io import get_file
    from .yamlutil import AsdfLoader, _IgnoreCustomTagsLoader

    if tagged:
        loader = AsdfLoader
    else:
        loader = _IgnoreCustomTagsLoader

    with get_file(init, "r") as gf:
        reader = gf.reader_until(
            constants.YAML_END_MARKER_REGEX,
            7,
            "End of YAML marker",
            include=True,
        )
        # The following call to yaml.load is safe because we're
        # using only loaders that don't create custom python objects
        content = yaml.load(reader, Loader=loader)  # noqa: S506
    return content


def get_array_base(arr):
    """
    For a given Numpy array, finds the base array that "owns" the
    actual data.
    """
    from .tags.core import ndarray

    base = arr
    while isinstance(base.base, (np.ndarray, ndarray.NDArrayType)):
        base = base.base
    return base


def get_base_uri(uri):
    """
    For a given URI, return the part without any fragment.
    """
    parts = _patched_urllib_parse.urlparse(uri)
    return _patched_urllib_parse.urlunparse([*list(parts[:5]), ""])


def _iter_subclasses(cls):
    """
    Returns all subclasses of a class.
    """
    for x in cls.__subclasses__():
        yield x
        yield from _iter_subclasses(x)


def calculate_padding(content_size, pad_blocks, block_size):
    """
    Calculates the amount of extra space to add to a block given the
    user's request for the amount of extra space.  Care is given so
    that the total of size of the block with padding is evenly
    divisible by block size.

    Parameters
    ----------
    content_size : int
        The size of the actual content

    pad_blocks : float or bool
        If `False`, add no padding (always return 0).  If `True`, add
        a default amount of padding of 10% If a float, it is a factor
        to multiple content_size by to get the new total size.

    block_size : int
        The filesystem block size to use.

    Returns
    -------
    nbytes : int
        The number of extra bytes to add for padding.
    """
    if not pad_blocks:
        return 0
    if pad_blocks is True:
        pad_blocks = 1.1
    new_size = content_size * pad_blocks
    new_size = int((math.ceil(float(new_size) / block_size) + 1) * block_size)
    return max(new_size - content_size, 0)


class _BinaryStruct:
    """
    A wrapper around the Python stdlib struct module to define a
    binary struct more like a dictionary than a tuple.
    """

    def __init__(self, descr, endian=">"):
        """
        Parameters
        ----------
        descr : list of tuple
            Each entry is a pair ``(name, format)``, where ``format``
            is one of the format types understood by `struct`.

        endian : str, optional
            The endianness of the struct.  Must be ``>`` or ``<``.
        """
        self._fmt = [endian]
        self._offsets = {}
        self._names = []
        i = 0
        for name, fmt in descr:
            self._fmt.append(fmt)
            self._offsets[name] = (i, (endian + fmt).encode("ascii"))
            self._names.append(name)
            i += struct.calcsize(fmt.encode("ascii"))
        self._fmt = "".join(self._fmt).encode("ascii")
        self._size = struct.calcsize(self._fmt)

    @property
    def size(self):
        """
        Return the size of the struct.
        """
        return self._size

    def pack(self, **kwargs):
        """
        Pack the given arguments, which are given as kwargs, and
        return the binary struct.
        """
        fields = [0] * len(self._names)
        for key, val in kwargs.items():
            if key not in self._offsets:
                msg = f"No header field '{key}'"
                raise KeyError(msg)
            i = self._names.index(key)
            fields[i] = val
        return struct.pack(self._fmt, *fields)

    def unpack(self, buff):
        """
        Unpack the given binary buffer into the fields.  The result
        is a dictionary mapping field names to values.
        """
        args = struct.unpack_from(self._fmt, buff[: self._size])
        return dict(zip(self._names, args))

    def update(self, fd, **kwargs):
        """
        Update part of the struct in-place.

        Parameters
        ----------
        fd : generic_io.GenericIO instance
            A writable, seekable file descriptor, currently seeked
            to the beginning of the struct.

        **kwargs : values
            The values to update on the struct.
        """
        updates = []
        for key, val in kwargs.items():
            if key not in self._offsets:
                msg = f"No header field '{key}'"
                raise KeyError(msg)
            updates.append((self._offsets[key], val))
        updates.sort()

        start = fd.tell()
        for (offset, datatype), val in updates:
            fd.seek(start + offset)
            fd.write(struct.pack(datatype, val))


class HashableDict(dict):
    """
    A simple wrapper around dict to make it hashable.

    This is sure to be slow, but for small dictionaries it shouldn't
    matter.
    """

    def __hash__(self):
        return hash(frozenset(self.items()))


def get_class_name(obj, instance=True):
    """
    Given a class or instance of a class, returns a string representing the
    fully specified path of the class.

    Parameters
    ----------

    obj : object
        An instance of any object
    instance: bool
        Indicates whether given object is an instance of the class to be named
    """
    typ = type(obj) if instance else obj
    return f"{typ.__module__}.{typ.__qualname__}"


class _InheritDocstrings(type):
    """
    This metaclass makes methods of a class automatically have their
    docstrings filled in from the methods they override in the base
    class.

    If the class uses multiple inheritance, the docstring will be
    chosen from the first class in the bases list, in the same way as
    methods are normally resolved in Python.  If this results in
    selecting the wrong docstring, the docstring will need to be
    explicitly included on the method.

    For example::

        >>> from asdf.util import _InheritDocstrings
        >>> class A(metaclass=_InheritDocstrings):
        ...     def wiggle(self):
        ...         "Wiggle the thingamajig"
        ...         pass
        >>> class B(A):
        ...     def wiggle(self):
        ...         pass
        >>> B.wiggle.__doc__
        'Wiggle the thingamajig'
    """

    def __init__(cls, name, bases, dct):
        def is_public_member(key):
            return (key.startswith("__") and key.endswith("__") and len(key) > 4) or not key.startswith("_")

        for key, val in dct.items():
            if inspect.isfunction(val) and is_public_member(key) and val.__doc__ is None:
                for base in cls.__mro__[1:]:
                    super_method = getattr(base, key, None)
                    if super_method is not None:
                        val.__doc__ = super_method.__doc__
                        break

        super().__init__(name, bases, dct)


class _NotSetType:
    def __repr__(self):
        return "NotSet"


"""
Special value indicating that a parameter is not set.  Distinct
from None, which may for example be a value of interest in a search.
"""
NotSet = _NotSetType()


def uri_match(pattern, uri):
    """
    Determine if a URI matches a URI pattern with possible
    wildcards.  The two recognized wildcards:

    "*":  match any character except /

    "**": match any character

    Parameters
    ----------
    pattern : str
        URI pattern.
    uri : str
        URI to check against the pattern.

    Returns
    -------
    bool
        `True` if URI matches the pattern.
    """
    if not isinstance(uri, str):
        return False

    if "*" in pattern:
        return _compile_uri_match_pattern(pattern).fullmatch(uri) is not None

    return pattern == uri


@lru_cache(1024)
def _compile_uri_match_pattern(pattern):
    # Escape the pattern in case it contains regex special characters
    # ('.' in particular is common in URIs) and then replace the
    # escaped asterisks with the appropriate regex matchers.
    pattern = re.escape(pattern)
    pattern = pattern.replace(r"\*\*", r".*")
    pattern = pattern.replace(r"\*", r"[^/]*")
    return re.compile(pattern)


def get_file_type(fd):
    """
    Determine the file type of an open GenericFile instance.

    Parameters
    ----------

    fd : ``asdf.generic_io.GenericFile``

    Returns
    -------

    `asdf.util.FileType`
    """
    if fd.peek(5) == constants.ASDF_MAGIC:
        return FileType.ASDF

    if fd.peek(6) == constants.FITS_MAGIC:
        return FileType.FITS

    return FileType.UNKNOWN


class FileType(enum.Enum):
    """
    Enum representing if a file is ASDF, FITS or UNKNOWN.
    """

    ASDF = 1
    FITS = 2
    UNKNOWN = 3



import contextlib
import os
import pathlib
import re
import tempfile

import numpy as np
import pytest

import asdf
from asdf.extension import ExtensionProxy, ManifestExtension
from asdf.resource import DirectoryResourceMapping


def test_info_module(capsys, tmp_path):
    tree = {
        "foo": 42,
        "bar": "hello",
        "baz": np.arange(20),
        "nested": {"woo": "hoo", "yee": "haw"},
        "long_line": "a" * 100,
    }
    af = asdf.AsdfFile(tree)

    def _assert_correct_info(node_or_path):
        asdf.info(node_or_path)
        captured = capsys.readouterr()
        assert "foo" in captured.out
        assert "bar" in captured.out
        assert "baz" in captured.out

    _assert_correct_info(af)
    _assert_correct_info(af.tree)

    tmpfile = str(tmp_path / "written.asdf")
    af.write_to(tmpfile)
    af.close()

    _assert_correct_info(tmpfile)
    _assert_correct_info(pathlib.Path(tmpfile))

    for i in range(1, 10):
        asdf.info(af, max_rows=i)
        lines = capsys.readouterr().out.strip().split("\n")
        assert len(lines) <= i

    asdf.info(af, max_cols=80)
    assert "(truncated)" in capsys.readouterr().out
    asdf.info(af, max_cols=None)
    captured = capsys.readouterr().out
    assert "(truncated)" not in captured
    assert "a" * 100 in captured

    asdf.info(af, show_values=True)
    assert "hello" in capsys.readouterr().out
    asdf.info(af, show_values=False)
    assert "hello" not in capsys.readouterr().out

    tree = {"foo": ["alpha", "bravo", "charlie", "delta", "eagle"]}
    af = asdf.AsdfFile(tree)
    asdf.info(af, max_rows=(None,))
    assert "alpha" not in capsys.readouterr().out
    for i in range(1, 5):
        asdf.info(af, max_rows=(None, i))
        captured = capsys.readouterr()
        for val in tree["foo"][0 : i - 1]:
            assert val in captured.out
        for val in tree["foo"][i - 1 :]:
            assert val not in captured.out


class ObjectWithInfoSupport:
    def __init__(self, clown="", the_meaning=0, anyof=None, allof=None, oneof=None, **kw):
        self._tag = "asdf://somewhere.org/asdf/tags/foo-1.0.0"
        self.clown = clown
        self.the_meaning_of_life_the_universe_and_everything = the_meaning
        self.anyof = anyof
        self.allof = allof
        self.oneof = oneof
        self.patt = {}
        for key in kw:
            if re.search("^S_", key):
                if not isinstance(kw[key], str):
                    msg = "S_ pattern object must be a string"
                    raise ValueError(msg)
                self.patt[key] = kw[key]
            if re.search("^I_", key):
                if not isinstance(kw[key], int):
                    msg = "I_ pattern object must be an int"
                    raise ValueError(msg)
                self.patt[key] = kw[key]

    def __asdf_traverse__(self):
        returnval = {
            "the_meaning_of_life_the_universe_and_everything": self.the_meaning_of_life_the_universe_and_everything,
            "clown": self.clown,
        }
        if self.anyof is not None:
            returnval["anyof_attribute"] = self.anyof
        if self.allof is not None:
            returnval["allof_attribute"] = self.allof
        if self.oneof is not None:
            returnval["oneof_attribute"] = self.oneof
        for key in self.patt:
            returnval[key] = self.patt[key]
        return returnval


class ObjectWithInfoSupport2:
    def __init__(self, attribute1="", attribute2=""):
        self._tag = "asdf://somewhere.org/asdf/tags/bar-1.0.0"
        self.attribute1 = attribute1
        self.attribute2 = attribute2

    def __asdf_traverse__(self):
        return {
            "attribute1": self.attribute1,
            "attribute2": self.attribute2,
        }


class ObjectWithInfoSupport3:
    def __init__(self, attribute_one="", attribute_two=""):
        self._tag = "asdf://somewhere.org/asdf/tags/drink-1.0.0"
        self.attribute_one = attribute_one
        self.attribute_two = attribute_two

    def __asdf_traverse__(self):
        return {
            "attributeOne": self.attribute_one,
            "attributeTwo": self.attribute_two,
        }


@contextlib.contextmanager
def manifest_extension(tmp_path):
    foo_manifest = """%YAML 1.1
---
id: asdf://somewhere.org/asdf/manifests/foo_manifest-1.0
extension_uri: asdf://somewhere.org/asdf/extensions/foo_manifest-1.0
asdf_standard_requirement:
  gte: 1.6.0
  lt: 2.0.0
tags:
  - tag_uri: asdf://somewhere.org/asdf/tags/foo-1.0.0
    schema_uri: asdf://somewhere.org/asdf/schemas/foo-1.0.0
    title: Foo title
    description: Foo description
  - tag_uri: asdf://somewhere.org/asdf/tags/bar-1.0.0
    schema_uri: asdf://somewhere.org/asdf/schemas/bar-1.0.0
    title: Bar title
    description: Bar Description
  - tag_uri: asdf://somewhere.org/asdf/tags/drink-1.0.0
    schema_uri: asdf://somewhere.org/asdf/schemas/drink-1.0.0
    title: Drink title
    description: Drink Description
...
"""
    foo_schema = """
%YAML 1.1
---
$schema: "asdf://stsci.edu/schemas/asdf/asdf-schema-1.1.0"
id: "asdf://somewhere.org/asdf/schemas/foo-1.0.0"

type: object
title: object with info support title
description: object with info support description
properties:
  the_meaning_of_life_the_universe_and_everything:
    title: Some silly title
    description: Some silly description
    type: integer
    archive_catalog:
      datatype: int
      destination: [ScienceCommon.silly]
  clown:
    title: clown name
    description: clown description
    type: string
    archive_catalog:
        datatype: str
        destination: [ScienceCommon.clown]
  anyof_attribute:
    title: anyOf example attribute
    description: anyOf description
    anyOf:
      - type: string
      - type: number
      - type: object
        properties:
          value:
            title: nested object in anyof example
            description: nested object description
            type: integer
          comment:
            title: comment for property
            description: comment description
            type: string
      - tag: asdf://somewhere.org/tags/bar-1.0.0
  oneof_attribute:
    title: oneOf example attribute
    description: oneOf description
    oneOf:
      - type: integer
        multipleOf: 5
      - type: integer
        multipleOf: 3
  allof_attribute:
    title: allOf example attribute
    description: allOf description
    allOf:
      - type: string
      - maxLength: 5
  patternProperties:
    "^S_":
      title: string pattern property
      type: string
    "^I_":
      title: integer pattern property
      type: integer
...
"""

    bar_schema = """
%YAML 1.1
---
$schema: "asdf://stsci.edu/schemas/asdf/asdf-schema-1.1.0"
id: "asdf://somewhere.org/asdf/schemas/bar-1.0.0"

type: object
title: object with info support 2 title
properties:
  attribute1:
    title: Attribute1 Title
    type: string
    archive_catalog:
      datatype: str
      destination: [ScienceCommon.attribute1]
  attribute2:
    title: Attribute2 Title
    type: string
    archive_catalog:
      datatype: str
      destination: [ScienceCommon.attribute2]
...
"""

    drink_schema = """
%YAML 1.1
---
$schema: "asdf://stsci.edu/schemas/asdf/asdf-schema-1.1.0"
id: "asdf://somewhere.org/asdf/schemas/drink-1.0.0"

type: object
title: object with info support 3 title
description: object description
allOf:
    - $ref: drink_ref-1.0.0
...
"""
    drink_ref_schema = """
%YAML 1.1
---
$schema: "asdf://stsci.edu/schemas/asdf/asdf-schema-1.1.0"
id: "asdf://somewhere.org/asdf/schemas/drink_ref-1.0.0"
properties:
  attributeOne:
    title: AttributeOne Title
    description: AttributeOne description
    type: string
    archive_catalog:
      datatype: str
      destination: [ScienceCommon.attributeOne]
  attributeTwo:
    allOf:
    - title: AttributeTwo Title
      description: AttributeTwo description
      type: string
      archive_catalog:
        datatype: str
        destination: [ScienceCommon.attributeTwo]
...
"""
    os.mkdir(tmp_path / "schemas")
    spath = tmp_path / "schemas" / "foo-1.0.0.yaml"
    with open(spath, "w") as fschema:
        fschema.write(foo_schema)
    spath = tmp_path / "schemas" / "bar-1.0.0.yaml"
    with open(spath, "w") as fschema:
        fschema.write(bar_schema)
    spath = tmp_path / "schemas" / "drink-1.0.0.yaml"
    with open(spath, "w") as fschema:
        fschema.write(drink_schema)
    spath = tmp_path / "schemas" / "drink_ref-1.0.0.yaml"
    with open(spath, "w") as fschema:
        fschema.write(drink_ref_schema)
    os.mkdir(tmp_path / "manifests")
    mpath = str(tmp_path / "manifests" / "foo_manifest-1.0.yaml")
    with open(mpath, "w") as fmanifest:
        fmanifest.write(foo_manifest)

    class FooConverter:
        tags = ["asdf://somewhere.org/asdf/tags/foo-1.0.0"]
        types = [ObjectWithInfoSupport]

        def to_yaml_tree(self, obj, tag, ctx):
            node = {
                "the_meaning_of_life_the_universe_and_everything": obj.the_meaning_of_life_the_universe_and_everything,
                "clown": obj.clown,
            }
            if obj.anyof is not None:
                node["anyof_attribute"] = obj.anyof
            if obj.oneof is not None:
                node["oneof_attribute"] = obj.oneof
            if obj.allof is not None:
                node["allof_attribute"] = obj.allof
            return node

        def from_yaml_tree(self, node, tag, ctx):
            return ObjectWithInfoSupport(
                the_meaning=node["the_meaning_of_life_the_universe_and_everything"],
                clown=node["clown"],
            )

    class BarConverter:
        tags = ["asdf://somewhere.org/asdf/tags/bar-1.0.0"]
        types = [ObjectWithInfoSupport2]

        def to_yaml_tree(self, obj, tag, ctx):
            return {
                "attribute1": obj.attribute1,
                "attribute2": obj.attribute2,
            }

        def from_yaml_tree(self, node, tag, ctx):
            return ObjectWithInfoSupport(attribute1="value1", attribute2="value2")

    class DrinkConverter:
        tags = ["asdf://somewhere.org/asdf/tags/drink-1.0.0"]
        types = [ObjectWithInfoSupport3]

        def to_yaml_tree(self, obj, tag, ctx):
            return {
                "attributeOne": obj.attribute_one,
                "attributeTwo": obj.attribute_two,
            }

        def from_yaml_tree(self, node, tag, ctx):
            return ObjectWithInfoSupport(attribute_one="value1", attribute_two="value2")

    converter1 = FooConverter()
    converter2 = BarConverter()
    converter3 = DrinkConverter()

    with asdf.config_context() as config:
        config.add_resource_mapping(
            DirectoryResourceMapping(str(tmp_path / "manifests"), "asdf://somewhere.org/asdf/manifests/"),
        )
        config.add_resource_mapping(
            DirectoryResourceMapping(str(tmp_path / "schemas"), "asdf://somewhere.org/asdf/schemas/"),
        )
        extension = ManifestExtension.from_uri(
            "asdf://somewhere.org/asdf/manifests/foo_manifest-1.0",
            converters=[converter1, converter2, converter3],
        )
        proxy = ExtensionProxy(extension)
        config.add_extension(proxy)
        yield config


def create_tree():
    return {
        "random": 3.14159,
        "object": ObjectWithInfoSupport(
            "Bozo",
            42,
            anyof=ObjectWithInfoSupport2("VAL1", "VAL2"),
            oneof=20,
            allof="good",
            S_example="beep",
            I_example=1,
        ),
        "list_of_stuff": [
            ObjectWithInfoSupport3("v1", "v2"),
            ObjectWithInfoSupport3("x1", "x2"),
        ],
    }


def test_schema_info_support(tmp_path):
    with manifest_extension(tmp_path):
        af = asdf.AsdfFile()
        af.tree = create_tree()

        assert af.schema_info("title") == {
            "list_of_stuff": [
                {
                    "attributeOne": {
                        "title": ("AttributeOne Title", "v1"),
                    },
                    "attributeTwo": {
                        "title": ("AttributeTwo Title", "v2"),
                    },
                    "title": ("object with info support 3 title", af.tree["list_of_stuff"][0]),
                },
                {
                    "attributeOne": {
                        "title": ("AttributeOne Title", "x1"),
                    },
                    "attributeTwo": {
                        "title": ("AttributeTwo Title", "x2"),
                    },
                    "title": ("object with info support 3 title", af.tree["list_of_stuff"][1]),
                },
            ],
            "object": {
                "I_example": {"title": ("integer pattern property", 1)},
                "S_example": {"title": ("string pattern property", "beep")},
                "allof_attribute": {"title": ("allOf example attribute", "good")},
                "anyof_attribute": {
                    "attribute1": {
                        "title": ("Attribute1 Title", "VAL1"),
                    },
                    "attribute2": {
                        "title": ("Attribute2 Title", "VAL2"),
                    },
                    "title": ("object with info support 2 title", af.tree["object"].anyof),
                },
                "clown": {"title": ("clown name", "Bozo")},
                "oneof_attribute": {"title": ("oneOf example attribute", 20)},
                "the_meaning_of_life_the_universe_and_everything": {"title": ("Some silly title", 42)},
                "title": ("object with info support title", af.tree["object"]),
            },
        }

        assert af.schema_info("archive_catalog") == {
            "list_of_stuff": [
                {
                    "attributeOne": {
                        "archive_catalog": ({"datatype": "str", "destination": ["ScienceCommon.attributeOne"]}, "v1"),
                    },
                    "attributeTwo": {
                        "archive_catalog": ({"datatype": "str", "destination": ["ScienceCommon.attributeTwo"]}, "v2"),
                    },
                },
                {
                    "attributeOne": {
                        "archive_catalog": ({"datatype": "str", "destination": ["ScienceCommon.attributeOne"]}, "x1"),
                    },
                    "attributeTwo": {
                        "archive_catalog": ({"datatype": "str", "destination": ["ScienceCommon.attributeTwo"]}, "x2"),
                    },
                },
            ],
            "object": {
                "anyof_attribute": {
                    "attribute1": {
                        "archive_catalog": ({"datatype": "str", "destination": ["ScienceCommon.attribute1"]}, "VAL1"),
                    },
                    "attribute2": {
                        "archive_catalog": ({"datatype": "str", "destination": ["ScienceCommon.attribute2"]}, "VAL2"),
                    },
                },
                "clown": {
                    "archive_catalog": ({"datatype": "str", "destination": ["ScienceCommon.clown"]}, "Bozo"),
                },
                "the_meaning_of_life_the_universe_and_everything": {
                    "archive_catalog": ({"datatype": "int", "destination": ["ScienceCommon.silly"]}, 42),
                },
            },
        }

        assert af.schema_info("archive_catalog", preserve_list=False) == {
            "list_of_stuff": {
                0: {
                    "attributeOne": {
                        "archive_catalog": ({"datatype": "str", "destination": ["ScienceCommon.attributeOne"]}, "v1"),
                    },
                    "attributeTwo": {
                        "archive_catalog": ({"datatype": "str", "destination": ["ScienceCommon.attributeTwo"]}, "v2"),
                    },
                },
                1: {
                    "attributeOne": {
                        "archive_catalog": ({"datatype": "str", "destination": ["ScienceCommon.attributeOne"]}, "x1"),
                    },
                    "attributeTwo": {
                        "archive_catalog": ({"datatype": "str", "destination": ["ScienceCommon.attributeTwo"]}, "x2"),
                    },
                },
            },
            "object": {
                "anyof_attribute": {
                    "attribute1": {
                        "archive_catalog": ({"datatype": "str", "destination": ["ScienceCommon.attribute1"]}, "VAL1"),
                    },
                    "attribute2": {
                        "archive_catalog": ({"datatype": "str", "destination": ["ScienceCommon.attribute2"]}, "VAL2"),
                    },
                },
                "clown": {
                    "archive_catalog": ({"datatype": "str", "destination": ["ScienceCommon.clown"]}, "Bozo"),
                },
                "the_meaning_of_life_the_universe_and_everything": {
                    "archive_catalog": ({"datatype": "int", "destination": ["ScienceCommon.silly"]}, 42),
                },
            },
        }

        assert af.schema_info("title", "list_of_stuff") == [
            {
                "attributeOne": {
                    "title": ("AttributeOne Title", "v1"),
                },
                "attributeTwo": {
                    "title": ("AttributeTwo Title", "v2"),
                },
                "title": ("object with info support 3 title", af.tree["list_of_stuff"][0]),
            },
            {
                "attributeOne": {
                    "title": ("AttributeOne Title", "x1"),
                },
                "attributeTwo": {
                    "title": ("AttributeTwo Title", "x2"),
                },
                "title": ("object with info support 3 title", af.tree["list_of_stuff"][1]),
            },
        ]

        assert af.schema_info("title", "object") == {
            "I_example": {"title": ("integer pattern property", 1)},
            "S_example": {"title": ("string pattern property", "beep")},
            "allof_attribute": {"title": ("allOf example attribute", "good")},
            "anyof_attribute": {
                "attribute1": {
                    "title": ("Attribute1 Title", "VAL1"),
                },
                "attribute2": {
                    "title": ("Attribute2 Title", "VAL2"),
                },
                "title": ("object with info support 2 title", af.tree["object"].anyof),
            },
            "clown": {"title": ("clown name", "Bozo")},
            "oneof_attribute": {"title": ("oneOf example attribute", 20)},
            "the_meaning_of_life_the_universe_and_everything": {"title": ("Some silly title", 42)},
            "title": ("object with info support title", af.tree["object"]),
        }

        assert af.schema_info("title", "object.anyof_attribute") == {
            "attribute1": {
                "title": ("Attribute1 Title", "VAL1"),
            },
            "attribute2": {
                "title": ("Attribute2 Title", "VAL2"),
            },
            "title": ("object with info support 2 title", af.tree["object"].anyof),
        }

        assert af.schema_info("title", "object.anyof_attribute.attribute2") == {
            "title": ("Attribute2 Title", "VAL2"),
        }

        # Test printing the schema_info
        assert af.schema_info("title", "object.anyof_attribute.attribute2").__repr__() == "{'title': Attribute2 Title}"

        assert af.schema_info("title", "object.anyof_attribute.attribute2.foo") is None

        assert af.schema_info() == {
            "list_of_stuff": [
                {
                    "attributeOne": {"description": ("AttributeOne description", "v1")},
                    "attributeTwo": {"description": ("AttributeTwo description", "v2")},
                    "description": ("object description", af.tree["list_of_stuff"][0]),
                },
                {
                    "attributeOne": {"description": ("AttributeOne description", "x1")},
                    "attributeTwo": {"description": ("AttributeTwo description", "x2")},
                    "description": ("object description", af.tree["list_of_stuff"][1]),
                },
            ],
            "object": {
                "allof_attribute": {
                    "description": ("allOf description", "good"),
                },
                "clown": {
                    "description": ("clown description", "Bozo"),
                },
                "description": ("object with info support description", af.tree["object"]),
                "oneof_attribute": {
                    "description": ("oneOf description", 20),
                },
                "the_meaning_of_life_the_universe_and_everything": {
                    "description": ("Some silly description", 42),
                },
            },
        }

        # Test using a search result
        search = af.search("clown")
        assert af.schema_info("description", search) == {
            "object": {
                "clown": {
                    "description": ("clown description", "Bozo"),
                },
                "description": ("object with info support description", af.tree["object"]),
            },
        }


def test_info_object_support(capsys, tmp_path):
    with manifest_extension(tmp_path):
        af = asdf.AsdfFile()
        af.tree = create_tree()
        af.info()

        captured = capsys.readouterr()

        assert "the_meaning_of_life_the_universe_and_everything" in captured.out
        assert "clown" in captured.out
        assert "42" in captured.out
        assert "Bozo" in captured.out
        assert "clown name" in captured.out
        assert "silly" in captured.out
        assert "info support 2" in captured.out
        assert "Attribute2 Title" in captured.out
        assert "allOf example attribute" in captured.out
        assert "oneOf example attribute" in captured.out
        assert "string pattern property" in captured.out
        assert "integer pattern property" in captured.out
        assert "AttributeOne" in captured.out
        assert "AttributeTwo" in captured.out


class RecursiveObjectWithInfoSupport:
    def __init__(self):
        self._tag = "asdf://somewhere.org/asdf/tags/bar-1.0.0"
        self.the_meaning = 42
        self.clown = "Bozo"
        self.recursive = None

    def __asdf_traverse__(self):
        return {"the_meaning": self.the_meaning, "clown": self.clown, "recursive": self.recursive}

    def __str__(self):
        return "rec ref"


def test_recursive_info_object_support(capsys, tmp_path):
    tempdir = pathlib.Path(tempfile.mkdtemp())
    with manifest_extension(tempdir):
        af = asdf.AsdfFile()

        recursive_obj = RecursiveObjectWithInfoSupport()
        recursive_obj.recursive = recursive_obj
        tree = {"random": 3.14159, "rtest": recursive_obj}
        af = asdf.AsdfFile()
        # we need to do this to avoid validation against the
        # manifest (generated in manifest_extension) which is
        # now supported with the default asdf standard 1.6.0
        # I'm not sure why the manifest has this restriction
        # and prior to switching to the default 1.6.0 was ignored
        # which allowed this test to pass.
        af._tree = tree
        af.info()
        captured = capsys.readouterr()
        assert "recursive reference" in captured.out


def test_search():
    tree = {"foo": 42, "bar": "hello", "baz": np.arange(20)}
    af = asdf.AsdfFile(tree)

    result = af.search("foo")
    assert result.node == 42

    result = af.search(type_="ndarray")
    assert (result.node == tree["baz"]).all()

    result = af.search(value="hello")
    assert result.node == "hello"


def test_info_str(capsys):
    class BadStr:
        def __str__(self):
            raise Exception()

    class NewlineStr:
        def __str__(self):
            return "a\nb"

    class CarriageReturnStr:
        def __str__(self):
            return "a\rb"

    class NiceStr:
        def __str__(self):
            return "nice"

    af = asdf.AsdfFile()
    af["a"] = BadStr()
    af["b"] = NewlineStr()
    af["c"] = CarriageReturnStr()
    af["d"] = NiceStr()
    af.info()
    captured = capsys.readouterr()
    assert "(BadStr)\n" in captured.out
    assert "(NewlineStr)\n" in captured.out
    assert "(CarriageReturnStr)\n" in captured.out
    assert "(NiceStr): nice\n" in captured.out


@pytest.mark.parametrize(
    "schema, expected",
    [
        ({"properties": {"foo": {"type": "object"}}}, {"type": "object"}),
        ({"allOf": [{"properties": {"foo": {"type": "object"}}}]}, {"type": "object"}),
        ({"oneOf": [{"properties": {"foo": {"type": "object"}}}]}, {"type": "object"}),
        ({"anyOf": [{"properties": {"foo": {"type": "object"}}}]}, {"type": "object"}),
    ],
)
def test_node_property(schema, expected):
    ni = asdf._node_info.NodeSchemaInfo.from_root_node("title", "root", {}, schema)
    assert ni.get_schema_for_property("foo") == expected


@pytest.mark.parametrize(
    "schema",
    [
        {"not": {"properties": {"foo": {"type": "object"}}}},
        {"properties": {"foo": {"type": "object"}}, "allOf": [{"properties": {"foo": {"type": "object"}}}]},
        {"properties": {"foo": {"type": "object"}}, "anyOf": [{"properties": {"foo": {"type": "object"}}}]},
        {"properties": {"foo": {"type": "object"}}, "oneOf": [{"properties": {"foo": {"type": "object"}}}]},
        {
            "allOf": [{"properties": {"foo": {"type": "object"}}}],
            "anyOf": [{"properties": {"foo": {"type": "object"}}}],
        },
        {
            "anyOf": [{"properties": {"foo": {"type": "object"}}}],
            "oneOf": [{"properties": {"foo": {"type": "object"}}}],
        },
        {
            "oneOf": [{"properties": {"foo": {"type": "object"}}}],
            "allOf": [{"properties": {"foo": {"type": "object"}}}],
        },
    ],
)
def test_node_property_error(schema):
    ni = asdf._node_info.NodeSchemaInfo.from_root_node("title", "root", {}, schema)
    assert ni.get_schema_for_property("foo") == {}


@pytest.mark.parametrize(
    "schema, expected",
    [
        ({"title": "foo"}, "foo"),
        ({"allOf": [{"title": "foo"}]}, "foo"),
        ({"oneOf": [{"title": "foo"}]}, "foo"),
        ({"anyOf": [{"title": "foo"}]}, "foo"),
        ({"not": {"title": "foo"}}, None),
        ({"allOf": [{"title": "foo"}, {"title": "bar"}]}, None),
        ({"oneOf": [{"title": "foo"}, {"title": "bar"}]}, None),
        ({"anyOf": [{"title": "foo"}, {"title": "bar"}]}, None),
        ({"allOf": [{"title": "foo"}, {"title": "bar"}]}, None),
    ],
)
def test_node_info(schema, expected):
    ni = asdf._node_info.NodeSchemaInfo.from_root_node("title", "root", {}, schema)
    assert ni.info == expected


def test_info_with_custom_extension(capsys):
    MY_TAG_URI = "asdf://somewhere.org/tags/foo-1.0.0"
    MY_SCHEMA_URI = "asdf://somewhere.org/tags/foo-1.0.0"

    schema_bytes = f"""%YAML 1.1
---
$schema: "http://stsci.edu/schemas/yaml-schema/draft-01"
id: {MY_SCHEMA_URI}
title: top_title
properties:
    foo:
        title: foo_title
        type: object
        properties:
            bar:
                title: bar_title
""".encode("ascii")

    class MyExtension:
        extension_uri = "asdf://somewhere.org/extensions/foo-1.0.0"
        tags = [
            asdf.extension.TagDefinition(
                MY_TAG_URI,
                schema_uris=[MY_SCHEMA_URI],
            )
        ]

    class FooThing:
        def __asdf_traverse__(self):
            return {"bar": 1}

    class Thing:
        _tag = MY_TAG_URI

        def __asdf_traverse__(self):
            return {"foo": FooThing()}

    with asdf.config_context() as cfg:
        cfg.add_resource_mapping({MY_SCHEMA_URI: schema_bytes})
        ext = MyExtension()
        af = asdf.AsdfFile({"t": Thing()}, extensions=[ext])
        af.info(max_cols=None)

    captured = capsys.readouterr()
    assert "top_title" in captured.out
    assert "foo_title" in captured.out
    assert "bar_title" in captured.out


def test_info_no_infinite_loop(capsys):
    """
    Providing a recursive list used to cause an
    infinite loop. Test this is not the case.
    """
    af = asdf.AsdfFile()
    af["l"] = []
    af["l"].append(af["l"])
    af.info()
    captured = capsys.readouterr()
    assert "recursive" in captured.out


def _get_block_output(capsys, af, *args, **kwargs) -> str:
    """Passes provided arguments to both `asdf.info` and `AsdfFile.info` and verifies the outputs match.

    Returns the unified captured output.
    """
    # Set max_rows to zero to ignore any changes in tree formatting when computing snapshots
    af.info(*args, max_rows=0, **kwargs)
    method_cap = capsys.readouterr()

    asdf.info(af, *args, max_rows=0, **kwargs)
    mod_cap = capsys.readouterr()

    assert method_cap.out == mod_cap.out, "AsdfFile.info and asdf.info outputs do not match"
    return method_cap.out


@pytest.mark.parametrize("filename", ["ndarray0.asdf", "ndarray2.asdf", "simple_inline_array0.asdf"])
def test_info_blocks_show(capsys, test_data_path, filename, snapshot):
    """Verify block output for files with different numbers and types of blocks."""

    file_path = test_data_path / filename
    with asdf.open(file_path) as af:
        out = _get_block_output(capsys, af, show_blocks=True)
        # Run `pytest --snapshot-update` to update stored snapshot
        assert out == snapshot


def test_info_blocks_hide(capsys, test_data_path, snapshot):
    """Verify no block output is shown by default when the file contains blocks."""

    file_path = test_data_path / "ndarray0.asdf"
    with asdf.open(file_path) as af:
        out = _get_block_output(capsys, af)
        # Run `pytest --snapshot-update` to update stored snapshot
        assert out == snapshot



"""
Contains commands for dealing with exploded and imploded forms.
"""

import os

import asdf
from asdf import AsdfFile

from .main import Command

__all__ = ["explode", "implode"]


class Implode(Command):
    @classmethod
    def setup_arguments(cls, subparsers):
        parser = subparsers.add_parser(
            "implode",
            help="Implode a ASDF file.",
            description="""Combine a ASDF file, where the data may be
            stored in multiple ASDF files, into a single ASDF
            file.""",
        )

        parser.add_argument("filename", nargs=1, help="""The ASDF file to implode.""")
        parser.add_argument(
            "--output",
            "-o",
            type=str,
            nargs="?",
            help="""The name of the output file.  If not provided, it
            will be the name of the input file with "_all"
            appended.""",
        )
        parser.add_argument(
            "--resolve-references",
            "-r",
            action="store_true",
            help="""Resolve all references and store them directly in
            the output file.""",
        )

        parser.set_defaults(func=cls.run)

        return parser

    @classmethod
    def run(cls, args):
        return implode(args.filename[0], args.output, args.resolve_references)


def implode(input_, output=None, resolve_references=False):
    """
    Implode a given ASDF file, which may reference external data, back
    into a single ASDF file.

    Parameters
    ----------
    input_ : str or file-like object
        The input file.

    output : str of file-like object
        The output file.

    resolve_references : bool, optional
        If `True` resolve all external references before saving.
    """
    if output is None:
        base, _ = os.path.splitext(input_)
        output = base + "_all" + ".asdf"
    with asdf.open(input_) as ff:
        ff2 = AsdfFile(ff)
        if resolve_references:
            ff2.resolve_references()
        ff2.write_to(output, all_array_storage="internal")


class Explode(Command):
    @classmethod
    def setup_arguments(cls, subparsers):
        parser = subparsers.add_parser(
            "explode",
            help="Explode a ASDF file.",
            description="""From a single ASDF file, create a set of
            ASDF files where each data block is stored in a separate
            file.""",
        )

        parser.add_argument("filename", nargs=1, help="""The ASDF file to explode.""")
        parser.add_argument(
            "--output",
            "-o",
            type=str,
            nargs="?",
            help="""The name of the output file.  If not provided, it
            will be the name of the input file with "_exploded"
            appended.""",
        )

        parser.set_defaults(func=cls.run)

        return parser

    @classmethod
    def run(cls, args):
        return explode(args.filename[0], args.output)


def explode(input_, output=None):
    """
    Explode a given ASDF file so each data block is in a separate
    file.

    Parameters
    ----------
    input_ : str or file-like object
        The input file.

    output : str of file-like object
        The output file.
    """
    if output is None:
        base, _ = os.path.splitext(input_)
        output = base + "_exploded" + ".asdf"
    with asdf.open(input_) as ff:
        ff.write_to(output, all_array_storage="external")



from asdf.extension import Converter


class ExternalArrayReferenceConverter(Converter):
    tags = ["tag:stsci.edu:asdf/core/externalarray-1.0.0"]
    types = ["asdf.tags.core.external_reference.ExternalArrayReference"]

    def to_yaml_tree(self, obj, tag, ctx):
        return {
            "fileuri": obj.fileuri,
            "target": obj.target,
            "datatype": obj.dtype,
            "shape": list(obj.shape),
        }

    def from_yaml_tree(self, node, tag, ctx):
        from asdf.tags.core import ExternalArrayReference

        return ExternalArrayReference(node["fileuri"], node["target"], node["datatype"], tuple(node["shape"]))



import os
import tempfile
import sys
import errno


if os.name == 'nt': # pragma: no cover
    import random
    import time

    _rename = lambda src, dst: False
    _rename_atomic = lambda src, dst: False

    import ctypes

    _MOVEFILE_REPLACE_EXISTING = 0x1
    _MOVEFILE_WRITE_THROUGH = 0x8
    _MoveFileEx = ctypes.windll.kernel32.MoveFileExW

    def _rename(src, dst):
        if not isinstance(src, str):
            src = str(src, sys.getfilesystemencoding())
        if not isinstance(dst, str):
            dst = str(dst, sys.getfilesystemencoding())
        if _rename_atomic(src, dst):
            return True
        retry = 0
        rv = False
        while not rv and retry < 100:
            rv = _MoveFileEx(src, dst, _MOVEFILE_REPLACE_EXISTING |
                                       _MOVEFILE_WRITE_THROUGH)
            if not rv:
                time.sleep(0.001)
                retry += 1
        return rv

    # new in Vista and Windows Server 2008
    _CreateTransaction = ctypes.windll.ktmw32.CreateTransaction
    _CommitTransaction = ctypes.windll.ktmw32.CommitTransaction
    _MoveFileTransacted = ctypes.windll.kernel32.MoveFileTransactedW
    _CloseHandle = ctypes.windll.kernel32.CloseHandle

    def _rename_atomic(src, dst):
        ta = _CreateTransaction(None, 0, 0, 0, 0, 1000, 'Atomic rename')
        if ta == -1:
            return False
        try:
            retry = 0
            rv = False
            while not rv and retry < 100:
                rv = _MoveFileTransacted(src, dst, None, None,
                                         _MOVEFILE_REPLACE_EXISTING |
                                         _MOVEFILE_WRITE_THROUGH, ta)
                if rv:
                    rv = _CommitTransaction(ta)
                    break
                else:
                    time.sleep(0.001)
                    retry += 1
            return rv
        finally:
            _CloseHandle(ta)

    def atomic_rename(src, dst):
        # Try atomic or pseudo-atomic rename
        if _rename(src, dst):
            return
        # Fall back to "move away and replace"
        try:
            os.rename(src, dst)
        except OSError as e:
            if e.errno != errno.EEXIST:
                raise
            old = "%s-%08x" % (dst, random.randint(0, sys.maxsize))
            os.rename(dst, old)
            os.rename(src, dst)
            try:
                os.unlink(old)
            except Exception:
                pass
else:
    atomic_rename = os.rename


class _AtomicWFile(object):
    """Helper class for :func:`atomic_open`."""

    def __init__(self, f, tmp_filename, filename):
        self._f = f
        self._tmp_filename = tmp_filename
        self._filename = filename

    def __getattr__(self, attr):
        return getattr(self._f, attr)

    def __enter__(self):
        return self

    @property
    def name(self):
        return self._filename

    def close(self):
        if self._f.closed:
            return
        self._f.close()
        atomic_rename(self._tmp_filename, self._filename)

    def __exit__(self, exc_type, exc_value, tb):
        if exc_type is None:
            self.close()
        else:
            self._f.close()
            try:
                os.remove(self._tmp_filename)
            except OSError:
                pass

    def __repr__(self):
        return '<%s %s%r, mode %r>' % (
            self.__class__.__name__,
            self._f.closed and 'closed ' or '',
            self._filename,
            self._f.mode
        )


def atomic_open(filename, mode='w'):
    """Works like a regular `open()` but writes updates into a temporary
    file instead of the given file and moves it over when the file is
    closed.  The file returned behaves as if it was a regular Python
    """
    if mode in ('r', 'rb', 'r+', 'rb+', 'a', 'ab'):
        raise TypeError('Read or append modes don\'t work with atomic_open')
    f = tempfile.NamedTemporaryFile(mode, prefix='.asdf_tmp',
                                    dir=os.path.dirname(filename),
                                    delete=False)
    return _AtomicWFile(f, f.name, filename)



"""
A `DataCallback` class is implemented here to allow
for reassignment of the index of an ASDF block corresponding
to a callback.

This is needed so that extension code can generate a callback
during deserialization of an ASDF file that will continue
to be valid even after an `AsdfFile.update` which might
reorder blocks.

To allow for 'low-level' block access needed for ndarray
`DataCallback` can be called with an optional ``_attr``
argument to cache data, access the block header and other
operations that we generally do not want to expose to
extension code.
"""

import weakref


class DataCallback:
    """
    A callable object used to read data from an ASDF block
    read from an ASDF file.
    """

    def __init__(self, index, read_blocks):
        self._reassign(index, read_blocks)

    def __call__(self, _attr=None):
        read_blocks = self._read_blocks_ref()
        if read_blocks is None:
            msg = "Attempt to read block data from missing block"
            raise OSError(msg)
        if _attr is None:
            return read_blocks[self._index].data
        else:
            # _attr allows NDArrayType to have low level block access for things
            # like reading the header and cached_data
            return getattr(read_blocks[self._index], _attr)

    def _reassign(self, index, read_blocks):
        self._index = index
        self._read_blocks_ref = weakref.ref(read_blocks)



import numpy as np

from asdf.extension import Converter


class NDArrayConverter(Converter):
    tags = [
        "tag:stsci.edu:asdf/core/ndarray-1.0.0",
        "tag:stsci.edu:asdf/core/ndarray-1.1.0",
        "tag:stsci.edu:asdf/core/ndarray-1.2.0",
    ]
    types = [
        np.ndarray,  # we use the type here so the extension can find the sub-classes
        "numpy.ma.MaskedArray",  # in numpy 2.0
        "numpy.ma.core.MaskedArray",
        "asdf.tags.core.ndarray.NDArrayType",
        "asdf.tags.core.stream.Stream",
    ]

    def to_yaml_tree(self, obj, tag, ctx):
        import numpy as np
        from numpy import ma

        from asdf import config, util
        from asdf._block.options import Options
        from asdf.tags.core.ndarray import NDArrayType, numpy_array_to_list, numpy_dtype_to_asdf_datatype
        from asdf.tags.core.stream import Stream

        data = obj

        if isinstance(obj, Stream):
            # previously, stream never passed on data, we can do that here
            ctx._blocks.set_streamed_write_block(data._array, data)

            result = {}
            result["source"] = -1
            result["shape"] = ["*", *data._shape]
            result["datatype"] = data._datatype
            result["byteorder"] = data._byteorder
            if data._strides is not None:
                result["strides"] = data._strides
            return result

        # sort out block writing options
        if isinstance(obj, NDArrayType) and isinstance(obj._source, str):
            # this is an external block, if we have no other settings, keep it as external
            options = ctx._blocks.options.lookup_by_object(data)
            if options is None:
                options = Options("external")
        else:
            options = ctx._blocks.options.get_options(data)

        # The ndarray-1.0.0 schema does not permit 0 valued strides.
        # Perhaps we'll want to allow this someday, to efficiently
        # represent an array of all the same value.
        if any(stride == 0 for stride in data.strides):
            data = np.ascontiguousarray(data)

        # Use the base array if that option is set or if the option
        # is unset and the AsdfConfig default is set
        cfg = config.get_config()
        if options.save_base or (options.save_base is None and cfg.default_array_save_base):
            base = util.get_array_base(data)
        else:
            base = data

        # The view computations that follow assume that the base array
        # is contiguous.  If not, we need to make a copy to avoid
        # writing a nonsense view.
        if not base.flags.forc:
            data = np.ascontiguousarray(data)
            base = util.get_array_base(data)

        shape = data.shape

        # possibly override options based on config settings
        if cfg.all_array_storage is not None:
            options.storage_type = cfg.all_array_storage
        if cfg.all_array_compression != "input":
            options.compression = cfg.all_array_compression
            options.compression_kwargs = cfg.all_array_compression_kwargs
        inline_threshold = cfg.array_inline_threshold

        if inline_threshold is not None and options.storage_type in ("inline", "internal"):
            if data.size < inline_threshold:
                options.storage_type = "inline"
            else:
                options.storage_type = "internal"
        ctx._blocks.options.set_options(data, options)

        # Compute the offset relative to the base array and not the
        # block data, in case the block is compressed.
        offset = data.ctypes.data - base.ctypes.data

        strides = None if data.flags.c_contiguous else data.strides
        dtype, byteorder = numpy_dtype_to_asdf_datatype(
            data.dtype,
            include_byteorder=(options.storage_type != "inline"),
        )

        result = {}

        result["shape"] = list(shape)
        if options.storage_type == "streamed":
            result["shape"][0] = "*"

        if options.storage_type == "inline":
            listdata = numpy_array_to_list(data)
            result["data"] = listdata
            result["datatype"] = dtype

        else:
            result["shape"] = list(shape)
            if options.storage_type == "streamed":
                result["shape"][0] = "*"
                result["source"] = -1
                ctx._blocks.set_streamed_write_block(base, data)
            else:
                result["source"] = ctx._blocks.make_write_block(base, options, obj)
            result["datatype"] = dtype
            result["byteorder"] = byteorder

            if offset > 0:
                result["offset"] = offset

            if strides is not None:
                result["strides"] = list(strides)

        if isinstance(data, ma.MaskedArray):
            if options.storage_type == "inline":
                ctx._blocks._set_array_storage(data.mask, "inline")

            result["mask"] = data.mask

        return result

    def from_yaml_tree(self, node, tag, ctx):
        import sys
        import weakref

        from asdf.tags.core import NDArrayType
        from asdf.tags.core.ndarray import asdf_datatype_to_numpy_dtype

        if isinstance(node, list):
            instance = NDArrayType(node, None, None, None, None, None, None)
            ctx._blocks._set_array_storage(instance, "inline")
            if not ctx._blocks._lazy_load:
                return instance._make_array()
            return instance

        if isinstance(node, dict):
            shape = node.get("shape", None)
            if "source" in node and "data" in node:
                msg = "Both source and data may not be provided at the same time"
                raise ValueError(msg)
            if "source" in node:
                source = node["source"]
                byteorder = node["byteorder"]
            else:
                source = node["data"]
                byteorder = sys.byteorder
            dtype = asdf_datatype_to_numpy_dtype(node["datatype"], byteorder) if "datatype" in node else None
            offset = node.get("offset", 0)
            strides = node.get("strides", None)
            mask = node.get("mask", None)

            if isinstance(source, int):
                # internal block
                data_callback = ctx.get_block_data_callback(source)
                instance = NDArrayType(source, shape, dtype, offset, strides, "A", mask, data_callback)
            elif isinstance(source, str):
                # external
                def data_callback(_attr=None, _ref=weakref.ref(ctx._blocks)):
                    if _attr not in (None, "cached_data", "data"):
                        raise AttributeError(f"_attr {_attr} is not supported")
                    blks = _ref()
                    if blks is None:
                        msg = "Failed to resolve reference to AsdfFile to read external block"
                        raise OSError(msg)
                    array = blks._load_external(source)
                    blks._set_array_storage(array, "external")
                    return array

                instance = NDArrayType(source, shape, dtype, offset, strides, "A", mask, data_callback)
            else:
                # inline
                instance = NDArrayType(source, shape, dtype, offset, strides, "A", mask)
                ctx._blocks._set_array_storage(instance, "inline")

            if not ctx._blocks._lazy_load:
                return instance._make_array()
            return instance

        msg = "Invalid ndarray description."
        raise TypeError(msg)

    def to_info(self, obj):
        return {"shape": obj.shape, "dtype": obj.dtype}



import itertools
import string

import numpy as np
import pytest

import asdf

SMALL_SET = string.ascii_lowercase[:3]
LARGE_SET = string.ascii_lowercase[:26]

DATA_GENS = ["no_data", "small_data", "large_data"]
TREE_GENS = ["small_tree", "flat_tree", "deep_tree", "large_tree"]
ALL_TREES = list(itertools.product(DATA_GENS, TREE_GENS))


def no_data(i):
    return i


def small_data(i):
    return np.full((3, 3), i)


def large_data(i):
    return np.full((128, 128), i)


def small_tree(data_gen):
    return {k: data_gen(ord(k)) for k in SMALL_SET}


def flat_tree(data_gen):
    return {k: data_gen(ord(k)) for k in LARGE_SET}


def deep_tree(data_gen):
    tree = {}
    for k in LARGE_SET:
        tree[k] = {"value": data_gen(k)}
        tree = tree[k]
    return tree


def large_tree(data_gen):
    tree = {}
    for k in LARGE_SET:
        tree["value"] = {k: data_gen(ord(k)) for k in LARGE_SET}
        tree[k] = {}
        tree = tree[k]
    return tree


@pytest.fixture(params=ALL_TREES, ids=["-".join(t) for t in ALL_TREES])
def tree(request):
    data_gen_name, tree_gen_name = request.param
    # some gymnastics to work around pytest not allowing
    # getfixturevalue to use parametrized fixtures
    return globals()[tree_gen_name](globals()[data_gen_name])


@pytest.fixture
def tree_bytes(tree):
    return asdf.dumps(tree)


@pytest.fixture
def asdf_file(tree):
    return asdf.AsdfFile(tree)



import http.server
import os
import queue
import shutil
import socketserver
import tempfile
import threading

__all__ = ["HTTPServer"]


def run_server(tmp_path, handler_class, stop_event, queue):  # pragma: no cover
    """
    Runs an HTTP server serving files from given tmp_path in a separate
    process.  When it's ready, it sends a URL to the server over a
    queue so the main process (the HTTP client) can start making
    requests of it.
    """

    class HTTPRequestHandler(handler_class):
        def translate_path(self, path):
            path = handler_class.translate_path(self, path)
            return os.path.join(tmp_path, os.path.relpath(path, os.getcwd()))

    server = socketserver.TCPServer(("127.0.0.1", 0), HTTPRequestHandler)
    domain, port = server.server_address
    url = f"http://{domain}:{port}/"

    # Set a reasonable timeout so that invalid requests (which may occur during
    # testing) do not cause the entire test suite to hang indefinitely
    server.timeout = 0.1

    queue.put(url)

    # Using server.serve_forever does not work here since it ignores the
    # timeout value set above. Having an explicit loop also allows us to kill
    # the server from the parent thread.
    while not stop_event.is_set():
        server.handle_request()

    server.server_close()


class HTTPServer:
    handler_class = http.server.SimpleHTTPRequestHandler

    def __init__(self):
        self.tmpdir = tempfile.mkdtemp()

        q = queue.Queue()
        self.stop_event = threading.Event()

        args = (self.tmpdir, self.handler_class, self.stop_event, q)
        self.thread = threading.Thread(target=run_server, args=args)
        self.thread.start()

        self.url = q.get()

    def finalize(self):
        self.stop_event.set()
        self.thread.join()
        shutil.rmtree(self.tmpdir)
